import numpy as np
import scipy.special as special

from waveforms import *


def test_waveform():
    t = np.linspace(-10, 10, 1001)

    wav = cos(1)
    assert np.allclose(wav(t), np.cos(t), atol=1e-04)

    wav.start = -10
    wav.stop = 10.02
    wav.sample_rate = 50
    assert np.allclose(wav.sample(), np.cos(t), atol=1e-04)

    wav = sin(1)
    assert np.allclose(wav(t), np.sin(t), atol=1e-04)

    width = 2
    wav = gaussian(width)
    std_sq2 = width / (4 * np.sqrt(np.log(2)))
    assert np.allclose(wav(t), np.exp(-(t / std_sq2)**2), atol=5e-03)

    wav = poly([1, -1 / 2, 1 / 6, -1 / 12])
    assert np.allclose(wav(t), np.poly1d([-1 / 12, 1 / 6, -1 / 2, 1])(t))

    sample_rate = 4e9
    width = 20e-9
    time_line = np.linspace(0, width * 100, int(width * 100 * sample_rate))
    wave = square(width) >> (width * 2)
    points = wave(time_line)
    assert isinstance(points, np.ndarray)


def test_tolist():
    pulse = gaussian(10) >> 5
    pulse += gaussian(10) >> 50
    pulse = pulse * cos(200)

    l = pulse.tolist()
    assert l == [
        np.inf, -np.inf, None, None, None, 5, -2.5, 0, 12.5, 1, 1.0, 2, 1, 3,
        2, 3.0028060219661246, 5, 1, 3, 4, 200, 0.0, 42.5, 0, 57.5, 1, 1.0, 2,
        1, 3, 2, 3.0028060219661246, 50, 1, 3, 4, 200, 0.0, np.inf, 0
    ]

    assert Waveform.fromlist(l) == pulse


def test_totree():
    pulse = gaussian(10) >> 5
    pulse += gaussian(10) >> 50
    pulse = pulse * cos(200)

    t = pulse.totree()
    assert t == ((np.inf, -np.inf, None, None, None),
                 ((-2.5, ()), (12.5, ((1.0, ((1, (2, 3.0028060219661246, 5)),
                                             (1, (4, 200, 0.0)))), )),
                  (42.5, ()), (57.5, ((1.0, ((1, (2, 3.0028060219661246, 50)),
                                             (1, (4, 200, 0.0)))), )), (np.inf,
                                                                        ())))
    assert Waveform.fromtree(t) == pulse


def test_op():
    t = np.linspace(-10, 10, 1001)

    wav = cos(1) + sin(2)
    assert np.allclose(wav(t), np.cos(t) + np.sin(2 * t))
    wav = cos(1) - sin(2)
    assert np.allclose(wav(t), np.cos(t) - np.sin(2 * t))
    wav = cos(1) * sin(2)
    assert np.allclose(wav(t), np.cos(t) * np.sin(2 * t))
    wav = cos(1) / 2
    assert np.allclose(wav(t), np.cos(t) / 2)


def test_simplify():
    t = np.linspace(-10, 10, 1001)
    wav = cos(1) * sin(2) * cos(3, 4)
    wav2 = wav.simplify()

    assert np.allclose(wav(t), np.cos(t) * np.sin(2 * t) * np.cos(3 * t + 4))
    assert np.allclose(wav2(t), np.cos(t) * np.sin(2 * t) * np.cos(3 * t + 4))

    wav = 2 * (exp(1.01 + 22j)**2 << 1) * exp(1.01 + 22j)
    wav2 = wav.simplify()

    assert np.allclose(
        wav(t), 2 * np.exp((1.01 + 22j) * (t + 1))**2 * np.exp(
            (1.01 + 22j) * t))
    assert np.allclose(wav2(t), wav(t))


def test_shift():
    t = np.linspace(-10, 10, 1001)
    width = 2
    wav = gaussian(width) >> 3
    std_sq2 = width / (4 * np.sqrt(np.log(2)))
    assert np.allclose(wav(t), np.exp(-((t - 3) / std_sq2)**2), atol=5e-03)


def test_chirp():
    t = np.linspace(0, 10, 1000, endpoint=False)

    def _chirp(t, f0, f1, T, phi0=0, type='linear'):
        if type == 'linear':
            return np.sin(phi0 + 2 * np.pi * ((f1 - f0) /
                                              (2 * T) * t**2 + f0 * t))
        elif type == 'exponential':
            return np.sin(phi0 + 2 * np.pi * f0 * T *
                          ((f1 / f0)**(t / T) - 1) / np.log((f1 / f0)))
        elif type == 'hyperbolic':
            return np.sin(phi0 - 2 * np.pi * f0 * f1 * T /
                          (f1 - f0) * np.log(1 - (f1 - f0) * t / (f1 * T)))
        else:
            raise ValueError(f'Unknow type {type}')

    wav1 = chirp(1, 2, 10, 4, 'linear')
    wav2 = chirp(1, 2, 10, 4, 'exponential')
    wav3 = chirp(1, 2, 10, 4, 'hyperbolic')

    assert np.allclose(wav1(t), _chirp(t, 1, 2, 10, 4, 'linear'))
    assert np.allclose(wav2(t), _chirp(t, 1, 2, 10, 4, 'exponential'))
    assert np.allclose(wav3(t), _chirp(t, 1, 2, 10, 4, 'hyperbolic'))


def test_parser():
    w1 = (gaussian(10) <<
          100) + square(20, edge=5, type='linear') * cos(2 * pi * 23.1)
    w2 = wave_eval(
        "(gaussian(10) << 100) + square(20, edge=5, type='linear') * cos(2*pi*23.1)"
    )
    w3 = wave_eval(
        "((gaussian(10) << 50) + ((square(20, 5, type='linear') * cos(2*pi*23.1)) >> 50)) << 50"
    )
    w4 = wave_eval(
        "(gaussian(10) << 100) + square(20, 5, 'linear') * cos(2*pi*23.1)")
    assert w1 == w2
    assert w1 == w3
    assert w1 == w4

    w1 = poly([1, -1 / 2, 1 / 6, -1 / 12])
    w2 = wave_eval("poly([1, -1/2, 1/6, -1/12])")
    w3 = wave_eval("poly((1, -1/2, 1/6, -1/12))")

    assert w1 == w2
    assert w1 == w3



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM qreg statement."""
import warnings

from .node import Node


class Qreg(Node):
    """Node for an OPENQASM qreg statement.
    children[0] is an indexedid node.
    """

    def __init__(self, children):
        """Create the qreg node."""
        super().__init__('qreg', children, None)
        # This is the indexed id, the full "id[n]" object
        self.id = children[0]  # pylint: disable=invalid-name
        # Name of the qreg
        self.name = self.id.name
        # Source line number
        self.line = self.id.line
        # Source file name
        self.file = self.id.file
        # Size of the register
        self.index = self.id.index

    def to_string(self, indent):
        """Print the node data, with indent."""
        ind = indent * ' '
        print(ind, 'qreg')
        self.children[0].to_string(indent + 3)

    def qasm(self, prec=None):
        """Return the corresponding OPENQASM string."""
        if prec is not None:
            warnings.warn('Parameter \'Qreg.qasm(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        return "qreg " + self.id.qasm() + ";"


from itertools import repeat
from typing import Sequence

import numpy as np
from scipy.fftpack import fft, fftfreq, ifft, ifftshift
from scipy.optimize import curve_fit
from scipy.signal import fftconvolve, lfilter


def shift(signal: np.ndarray, delay: float, dt: float) -> np.ndarray:
    """
    delay a signal

    Args:
        signal (np.ndarray): input signal
        delay (float): delayed time
        dt (float): time step of signal samples

    Returns:
        np.ndarray: delayed signal
    """
    points = int(delay // dt)
    delta = delay / dt - points

    if delta > 0:
        ker = np.array([0, 1 - delta, delta])
        signal = np.convolve(signal, ker, mode='same')

    if points == 0:
        return signal

    ret = np.zeros_like(signal)
    if points < 0:
        ret[:points] = signal[-points:]
    else:
        ret[points:] = signal[:-points]
    return ret


def extractKernel(sig_in, sig_out, sample_rate, bw=None, skip=0):
    corr = fft(sig_in) / fft(sig_out)
    ker = np.real(ifftshift(ifft(corr)))
    if bw is not None and bw < 0.5 * sample_rate:
        k = np.exp(-0.5 * np.linspace(-3.0, 3.0, int(2 * sample_rate / bw))**2)
        ker = np.convolve(ker, k / k.sum(), mode='same')
    return ker[int(skip):len(ker) - int(skip)]


def zDistortKernel(dt: float, params: Sequence[tuple]) -> np.ndarray:
    t = 3 * np.asarray(params)[:, 0].max()
    omega = 2 * np.pi * fftfreq(int(t / dt) + 1, dt)

    H = 1
    for tau, A in params:
        H += (1j * A * omega * tau) / (1j * omega * tau + 1)

    ker = ifftshift(ifft(1 / H)).real
    return ker


def exp_decay_filter_old(amp, tau, sample_rate):
    """
    exp decay filter

                    A
    H(w) = --------------------
            1 - 1j / (w * tau)

    Args:
        amp (float): amplitude of the filter
        tau (float): decay time
        sample_rate (float): sampling rate
    """

    alpha = 1 - np.exp(-1 / (abs(sample_rate * tau) * (1 + amp)))

    if amp >= 0:
        k = amp / (1 + amp - alpha)
        a = [(1 - k + k * alpha), -(1 - k) * (1 - alpha)]
    else:
        k = -amp / (1 + amp) / (1 - alpha)
        a = [(1 + k - k * alpha), -(1 + k) * (1 - alpha)]

    b = [1 / a[0], -(1 - alpha) / a[0]]
    a = [1, a[1] / a[0]]

    return b, a


def exp_decay_filter(amp: float | Sequence[float],
                     tau: float | Sequence[float],
                     sample_rate: float) -> tuple[np.ndarray, np.ndarray]:
    """
    exp decay filter

    Infinite impulse response as multiexponential decay. When input signal
    is the Heaviside theta function u(t), the output signal is:
    out(t) = u(t) * (1 - A_1 * exp(-t / tau_1) - A_2 * exp(-t / tau_2) ...)
    where A_i and tau_i are the amplitude and decay time of the i-th
    exponential decay.

    The transfer function of the filter is:

    H(w) = 1 - H_1(w) - H_2(w) - ... - H_n(w)

    where
                       A_i
    H_i(w) = --------------------------
              1 - 1 / (1j * w * tau_i)

    Args:
        amp (float): amplitude of the filter
        tau (float): decay time
        sample_rate (float): sampling rate

    Returns:
        tuple: (b, a) array like, numerator (b) and denominator (a)
        polynomials of the IIR filter. See scipy.signal.lfilter for more.
    """

    if isinstance(amp, (int, float, complex)):
        amp = [amp]
        tau = [tau]
    numerator, denominator = np.poly1d([0.0]), np.poly1d([1.0])
    for i, (A, t) in enumerate(zip(amp, tau)):
        denominator = denominator * np.poly1d([1, -1 / t])
        n = np.poly1d([-A, 0.0])
        for j, t_ in enumerate(tau):
            if j != i:
                n = n * np.poly1d([1, -1 / t_])
        numerator = numerator + n
    numerator = numerator + denominator
    xi = numerator.roots
    p = denominator.roots

    b, a = np.poly1d([1.0]), np.poly1d([1.0])
    for x in xi:
        b = b * np.poly1d([1, -np.exp(-x / sample_rate)])
    for p_ in p:
        a = a * np.poly1d([1, -np.exp(-p_ / sample_rate)])

    kd = numerator(0) * a(1) / denominator(0) / b(1)
    b, a = b.coeffs.real * kd, a.coeffs.real
    return b / a[0], a / a[0]


def reflection_filter(f, A, tau):
    """
    reflection filter

    Infinite impulse response as reflection. When input signal
    is in(t), the output signal is:
    out(t) = in(t) + A * in(t - tau) + A^2 * in(t - 2 * tau) + ...

    The transfer function of the filter is:
                      1 - A
    H(w) = ----------------------------
            1 - A * exp(- i * w * tau)
    Args:
        f (float): frequency
        A (float): amplitude of the reflection
        tau (float): delay time
    """
    return (1 - A) / (1 - A * np.exp(-2j * np.pi * f * tau))


def reflection(sig, A, tau, sample_rate):
    freq = np.fft.fftfreq(len(sig), 1 / sample_rate)
    return np.fft.ifft(np.fft.fft(sig) * reflection_filter(freq, A, tau)).real


def correct_reflection(sig, A, tau, sample_rate=None):
    from ...waveform import Waveform

    if isinstance(sig, Waveform):
        return 1 / (1 - A) * sig - A / (1 - A) * (sig >> tau)
    if sample_rate is not None:
        freq = np.fft.fftfreq(len(sig), 1 / sample_rate)
        return np.fft.ifft(np.fft.fft(sig) /
                           reflection_filter(freq, A, tau)).real
    else:
        raise ValueError('sample_rate is not given')


def predistort(sig: np.ndarray,
               filters: list = None,
               ker: np.ndarray = None) -> np.ndarray:
    if filters is None:
        filters = []
    for b, a in filters:
        sig = lfilter(b, a, sig)

    if ker is None:
        return sig

    size = len(sig)
    sig = np.hstack((np.zeros_like(sig), sig, np.zeros_like(sig)))
    start = size + len(ker) // 2
    stop = start + size
    return fftconvolve(sig, ker, mode='full')[start:stop]


def distort(points, params, sample_rate):
    filters = []
    for amp, tau in np.asarray(params).reshape(-1, 2):
        b, a = exp_decay_filter(amp, abs(tau), sample_rate)
        filters.append((b, a))
    return predistort(points, filters)


def phase_curve(t, params, df_dphi, pulse_width, start, wav, sample_rate):
    lim = max(np.max(np.abs(t)), 20e-6)
    num = round(2 * lim * sample_rate)
    tlist = np.arange(num) / sample_rate - lim
    points = wav(tlist)

    pulse_points = round(pulse_width * sample_rate)
    start_points = round((start + pulse_width) * sample_rate) - 1

    ker = np.hstack(
        [np.ones(pulse_points) / sample_rate,
         np.zeros(start_points)])

    points = np.convolve(2 * np.pi * df_dphi *
                         distort(points, params, sample_rate),
                         ker,
                         mode='same')
    return np.interp(t, tlist, points)


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from waveforms import square

    data = np.load('Z_distortion.npz')

    x = data['time'] * 1e-6
    y = data['phase']
    df_dphi = 4343.313e6

    sample_rate = 2e9
    wav = 0.1 * (square(2e-6) << 1e-6)

    def f(t, *params):
        return phase_curve(t, params, df_dphi, 10e-9, 25e-9)

    params = [-0.03, 0.1e-6, 0.02, 0.3e-6]
    popt, pcov = curve_fit(f, x, y, p0=params)

    plt.plot(x / 1e-6, y, 'o')
    plt.semilogx(
        x / 1e-6,
        phase_curve(x,
                    params,
                    df_dphi,
                    10e-9,
                    0,
                    wav=wav,
                    sample_rate=sample_rate))
    plt.plot(x / 1e-6, f(x, *popt))

    plt.xlabel('delay [us]')
    plt.ylabel('phase')
    plt.show()



import operator

from waveforms.qlisp.interpreter import OPCODE, VirtualMachine, compile, link

result = None


def output(arg):
    global result
    result = arg


#!yapf: disable
prog1 = [
    1., 4.,
    2, ':sub', OPCODE.CALL,
    1, output, OPCODE.CALL_RET
]

prog2 = [
    5,  # limit
    0,  # ret
    0,  # i

    'label:startloop',
    0, 0, OPCODE.SLOAD,     # load limit
    ':end', OPCODE.JE,      # if limit == 0 goto end

    1, 0, OPCODE.SLOAD,     # load ret
    2, 0, OPCODE.SLOAD,     # load i
    2, operator.add, OPCODE.CALL,
    1, 0, OPCODE.SSTORE,    # set ret

    2, 0, OPCODE.SLOAD,
    1,
    2, operator.add, OPCODE.CALL,
    2, 0, OPCODE.SSTORE,

    1,
    0, 0, OPCODE.SLOAD,
    2, operator.sub, OPCODE.CALL,
    0, 0, OPCODE.SSTORE,

    ':startloop', OPCODE.JMP,

    'label:end',

    1, 0, OPCODE.SLOAD,     # load ret
    1, output, OPCODE.CALL_RET
]

prog3 = [
    5,
    1, ':frac', OPCODE.CALL,
    1, output, OPCODE.CALL_RET
]

prog4 = [
    15,
    1,
    ':feb',
    OPCODE.CALL,
    1,
    output,
    OPCODE.CALL_RET,
]

sub = [
    2, operator.sub, OPCODE.CALL_RET
]

frac = [
    OPCODE.DUP,
    ':ret0', OPCODE.JE,
    ':retn', OPCODE.JMP,

    'label:ret0',
    1, OPCODE.RET,

    'label:retn',
    OPCODE.DUP,
    1,
    OPCODE.SWAP,
    2, operator.sub, OPCODE.CALL,
    1, ':frac', OPCODE.CALL,
    2, operator.mul, OPCODE.CALL_RET
]

feb = [
    OPCODE.DUP,
    ':ret0', OPCODE.JE,

    OPCODE.DUP, 1, OPCODE.SWAP, 2, ":sub", OPCODE.CALL,
    ':ret1', OPCODE.JE,
    ':retn', OPCODE.JMP,

    'label:ret0',
    0, OPCODE.RET,

    'label:ret1',
    1, OPCODE.RET,

    'label:retn',
    OPCODE.DUP, 1, OPCODE.SWAP,
    2, ":sub", OPCODE.CALL,
    1, ':feb', OPCODE.CALL,
    OPCODE.SWAP,

    2, OPCODE.SWAP,
    2, ":sub", OPCODE.CALL,
    1, ':feb', OPCODE.CALL,
    2, operator.add, OPCODE.CALL_RET,
]
#!yapf: enable

functions = {
    "sub": sub,
    "frac": frac,
    "feb": feb,
}


def test_VM1():
    vm = VirtualMachine(debug=False)
    functions['main'] = prog1
    code = link(functions)
    vm.run(code)
    assert result == 3.0


def test_VM2():
    vm = VirtualMachine(debug=False)
    functions['main'] = prog2
    code = link(functions)
    vm.run(code)
    assert result == 10


def test_VM3():
    vm = VirtualMachine(debug=False)
    functions['main'] = prog3
    code = link(functions)
    vm.run(code)
    assert result == 120


def test_VM4():
    vm = VirtualMachine(debug=False)
    functions['main'] = prog4
    code = link(functions)
    vm.run(code)
    assert result == 610


prog5 = """
(output
((lambda (a b c d n t) (begin
    (while (!= n 0)
    (begin
        (if (== (% n 2) 1)
        (begin
            (setq t (+ (* a c) (+ (* b c) (* a d))))
            (setq b (+ (* a c) (* b d)))
            (setq a t))
        ())
    (setq n (>> n 1))
    (setq t (+ (* c c) (+ (* d c) (* c d))))
    (setq d (+ (* c c) (* d d)))
    (setq c t)))
(+ a b)))
1 -1 1 0 (cast_int {}) 0))
"""

prog6 = """
(begin

(define f1 (lambda (a b c d) (+ (* a c) (+ (* b c) (* a d)))))
(define f2 (lambda (a b c d) (+ (* a c) (* b d))))

(output
((lambda (a b c d n t) (begin
    (while (!= n 0)
    (begin
        (if (== (% n 2) 1)
        (begin
            (setq t (f1 a b c d))
            (setq b (f2 a b c d))
            (setq a t))
        ())
    (setq n (>> n 1))
    (setq t (f1 c d c d))
    (setq d (f2 c d c d))
    (setq c t)))
(+ a b)))
1 -1 1 0 (cast_int {}) 0)))
"""

prog7 = """
(begin

(define frac (lambda (n) (if (== n 0) 1 (* n (frac (- n 1))))))

(output (frac {})))
"""

prog8 = """
(begin
    (define frac (lambda (n ret)
                         (if (== n 0)
                             ret
                             (frac (- n 1) (* ret n)))))
    (output (frac {} 1)))
"""


def feb(n):

    def feb_iter(a, b, c, d):
        return a * c + b * c + a * d, a * c + b * d

    a, b = 1, -1
    c, d = 1, 0
    while n != 0:
        if n % 2 == 1:
            a, b = feb_iter(a, b, c, d)
        n >>= 1
        c, d = feb_iter(c, d, c, d)
    return a + b


def frac(n):
    ret = 1
    while n != 0:
        ret = ret * n
        n = n - 1
    return ret


def test_comiler1():
    code = compile(prog5.format(10), extra_commands={'output': output})
    vm = VirtualMachine(debug=False)
    vm.run(code)
    assert result == feb(10)


def test_comiler2():
    code = compile(prog6.format(10), extra_commands={'output': output})
    vm = VirtualMachine(debug=False)
    vm.run(code)
    assert result == feb(10)


def test_comiler3():
    code = compile(prog7.format(10), extra_commands={'output': output})
    vm = VirtualMachine(debug=False)
    vm.run(code)
    assert result == frac(10)


def test_comiler4():
    code = compile(prog8.format(100), extra_commands={'output': output})
    vm = VirtualMachine(debug=False)
    vm.run(code)
    assert result == frac(100)
    assert len(vm.stack) < 20



import numpy as np
from scipy.optimize import curve_fit
from scipy.signal import find_peaks


def fit_rabi(x, y, Tr=100e-6, phi=-np.pi, freq=None):

    def func(t, A, Tr, freq, phi, B):
        return A * np.exp(-t / Tr) * np.cos(2 * np.pi * freq * t + phi) + B

    A = (np.max(y) - np.min(y)) / 2
    B = (np.max(y) + np.min(y)) / 2
    if freq is not None:
        p0 = [A, Tr, freq, phi, B]
    else:
        freq = np.fft.fftfreq(x.shape[0], x[1] - x[0])[1:(x.shape[0] // 2)]
        amp = np.abs(np.fft.fft(y))[1:(x.shape[0] // 2)]
        index, peak_height = find_peaks(amp / np.max(amp),
                                        height=0.01,
                                        distance=(x.shape[0] // 2))
        index = index[np.argsort(peak_height)[-1]]
        freq = freq[index]
        p0 = [A, Tr, freq, phi, B]
    try:
        popt, pcov = curve_fit(func,
                               x,
                               y,
                               p0=p0,
                               method='trf',
                               bounds=((max(A - 0.3,
                                            0), 10e-9, freq * 0.4, -np.pi,
                                        B - 0.3), (A + 0.3, np.inf, freq * 1.6,
                                                   np.pi, B + 0.3)))
        fitted = func(x, *popt)
        return popt, pcov, fitted
    except:
        raise ValueError('TimeRabi fitting failed!')


def fit_ramsey(x, y, T1=1.0, Tphi=1.0, Delta=0, method='default'):
    pass



from __future__ import annotations

import operator
from collections import defaultdict
from enum import Enum, auto
from math import e, pi
from typing import Any

from .parse import parse as qlisp_parser
from .tokenize import Expression, Symbol


class OPCODE(Enum):
    NOP = auto()
    DUP = auto()
    DROP = auto()
    SWAP = auto()
    OVER = auto()
    SLOAD = auto()
    SSTORE = auto()
    LOAD = auto()
    STORE = auto()
    CALL = auto()
    RET = auto()
    CALL_RET = auto()
    JMP = auto()
    JNE = auto()
    JE = auto()
    EXIT = auto()

    # Waveform operators
    PLAY = auto()
    CAPTURE = auto()
    DELAY = auto()
    SET_PHASE = auto()
    ADD_PHASE = auto()
    SET_TIME = auto()
    BARRIER = auto()

    # Register commands
    _SP = auto()
    _BP = auto()
    _SL = auto()
    _PC = auto()
    _WRITE_SP = auto()
    _WRITE_BP = auto()
    _WRITE_SL = auto()
    _WRITE_PC = auto()

    def __repr__(self):
        return self.name


class Frame():

    def __init__(self, frequency: int = 0):
        self.frequency = frequency
        self.phase = 0
        self.time = 0


class Channel():

    def __init__(self, name: str = 'demo', group: str = 'default'):
        self.name = name
        self.group = group
        self.frames = defaultdict(Frame)


def barrier(*channels: Channel):
    t_max = 0
    for channel in channels:
        for frame in channel.frames.values():
            t_max = max(t_max, frame.time)
    for channel in channels:
        for frame in channel.frames.values():
            frame.time = t_max


def vm_pop(vm):
    vm._pop()


def vm_dup(vm):
    x = vm._pop()
    vm._push(x)
    vm._push(x)


def vm_swap(vm):
    v2 = vm._pop()
    v1 = vm._pop()
    vm._push(v2)
    vm._push(v1)


def vm_over(vm):
    x = vm._pop()
    y = vm._pop()
    vm._push(y)
    vm._push(x)
    vm._push(y)


def vm_sload(vm):
    level = vm._pop()
    n = vm._pop()
    addr = vm.bp
    for _ in range(level):
        addr = vm.stack[addr - 3]
    vm._push(vm.stack[addr + n])


def vm_sstore(vm):
    level = vm._pop()
    n = vm._pop()
    value = vm._pop()

    addr = vm.bp
    for _ in range(level):
        addr = vm.stack[addr - 3]
    vm.stack[addr + n] = value


def vm_load(vm):
    addr = vm._pop()
    vm._push(vm.mem[addr])


def vm_store(vm):
    addr = vm._pop()
    value = vm._pop()
    vm.mem[addr] = value


def vm_call(vm):
    func = vm._pop()
    argc = vm._pop()
    args = [vm._pop() for _ in range(argc)]

    if callable(func):
        vm._push(func(*args))
    elif isinstance(func, int):
        vm._push(vm.bp)
        vm._push(vm.sl)
        vm._push(vm.pc)
        vm.bp = vm.sp
        for arg in reversed(args):
            vm._push(arg)
        vm.sl = func
        vm.pc = func
    else:
        raise RuntimeError(f"not callable {func}")


def vm_ret(vm):
    result = vm._pop()
    vm.sp = vm.bp - 3
    vm.bp, vm.sl, vm.pc = vm.stack[vm.bp - 3:vm.bp]
    vm._push(result)


def vm_call_ret(vm):
    func = vm._pop()
    argc = vm._pop()

    if callable(func):
        args = reversed(vm.stack[vm.sp - argc:vm.sp])
        vm.sp = vm.bp - 3
        vm.bp, vm.sl, vm.pc = vm.stack[vm.bp - 3:vm.bp]
        vm._push(func(*args))
    elif isinstance(func, int):
        if vm.sp != vm.bp + argc:
            vm.stack[vm.bp:vm.bp + argc] = vm.stack[vm.sp - argc:vm.sp]
            vm.sp = vm.bp + argc
        vm.sl = func
        vm.pc = func
    else:
        raise RuntimeError(f"not callable {func}")


def vm_jmp(vm):
    vm.pc = vm._pop() + vm.sl


def vm_jne(vm):
    addr = vm._pop()
    cond = vm._pop()
    if cond:
        vm.pc = addr + vm.sl


def vm_je(vm):
    addr = vm._pop()
    cond = vm._pop()
    if cond == 0:
        vm.pc = addr + vm.sl


dispatch_table = {
    OPCODE.DUP: vm_dup,
    OPCODE.DROP: vm_pop,
    OPCODE.SWAP: vm_swap,
    OPCODE.OVER: vm_over,
    OPCODE.SLOAD: vm_sload,
    OPCODE.SSTORE: vm_sstore,
    OPCODE.LOAD: vm_load,
    OPCODE.STORE: vm_store,
    OPCODE.CALL: vm_call,
    OPCODE.RET: vm_ret,
    OPCODE.CALL_RET: vm_call_ret,
    OPCODE.JMP: vm_jmp,
    OPCODE.JNE: vm_jne,
    OPCODE.JE: vm_je,
}


class VirtualMachine:

    def __init__(self, debug=False, dispatch=dispatch_table):
        self.mem = []
        self.stack = []
        self.channels = defaultdict(Channel)
        self.sp = 0  # stack pointer
        self.bp = 0  # base pointer
        self.sl = 0  # static link
        self.pc = 0  # program counter
        self.clk = 0
        self.debug = debug
        self.dispatch = dispatch

        self._dispatch_register = {
            OPCODE._SP: 'sp',
            OPCODE._BP: 'bp',
            OPCODE._SL: 'sl',
            OPCODE._PC: 'pc',
            OPCODE._WRITE_SP: 'sp',
            OPCODE._WRITE_BP: 'bp',
            OPCODE._WRITE_SL: 'sl',
            OPCODE._WRITE_PC: 'pc',
        }

    def _next(self):
        self.pc += 1
        return self.mem[self.pc - 1]

    def _pop(self):
        self.sp -= 1
        return self.stack[self.sp]

    def _push(self, value):
        if len(self.stack) > self.sp:
            self.stack[self.sp] = value
        else:
            self.stack.append(value)
        self.sp += 1

    def _pick(self, n=0):
        return self.stack[self.sp - n - 1]

    def _play(self):
        frame = self._pop()
        pulse = self._pop()

    def _captrue(self):
        frame = self._pop()
        cbit = self._pop()

    def display(self):
        if not self.debug:
            return
        print(f'State[{self.clk}] ====================')
        print(f'      OP: ', self.mem[self.pc])
        print(f'   STACK: ', self.stack[:self.sp])
        print(f'      BP: ', self.bp)
        print(f'      SL: ', self.sl)
        print(f'      PC: ', self.pc)
        print('')

    def trace(self):
        if not self.debug:
            return
        self.display()

    def run(self, code, step_limit=-1):
        if len(code) > len(self.mem):
            self.mem.extend([0] * (len(code) - len(self.mem)))
        for i, c in enumerate(code):
            self.mem[i] = c
        self.sp = 0  # stack pointer
        self.bp = 0  # base pointer
        self.sl = 0  # static link
        self.pc = 0  # program counter
        self.clk = 0  # clock

        while True:
            self.trace()
            op = self._next()
            if self.clk == step_limit:
                break
            self.clk += 1
            if isinstance(op, OPCODE):
                if op in self.dispatch:
                    self.dispatch[op](self)
                elif op == OPCODE.NOP:
                    continue
                elif op == OPCODE.EXIT:
                    break
                elif op in [OPCODE._SP, OPCODE._BP, OPCODE._SL, OPCODE._PC]:
                    self._push(getattr(self, self._dispatch_register[op]))
                elif op in [
                        OPCODE._WRITE_SP, OPCODE._WRITE_BP, OPCODE._WRITE_SL,
                        OPCODE._WRITE_PC
                ]:
                    setattr(self, self._dispatch_register[op], self._pop())
                else:
                    raise RuntimeError(f"unknown command {op}")
            else:
                self._push(op)


class Context:

    def __init__(self):
        self.internal = {
            '+': operator.add,
            '-': operator.sub,
            '*': operator.mul,
            '/': operator.truediv,
            '//': operator.floordiv,
            '%': operator.mod,
            '==': operator.eq,
            '!=': operator.ne,
            '<': operator.lt,
            '>': operator.gt,
            '<=': operator.le,
            '>=': operator.ge,
            'and': operator.and_,
            'or': operator.or_,
            'not': operator.not_,
            '>>': operator.rshift,
            '<<': operator.lshift,
            '|': operator.or_,
            '&': operator.and_,
            '^': operator.xor,
            '~': operator.invert,
            'print': print,
            'input': input,
            'cast_int': int,
            'cast_float': float,
        }
        self.functions = {}
        self.constants = {}
        self.counter = 0
        self.env = {}
        self.outer = None
        self.namespaces = []

    def new_label(self, label):
        self.counter += 1
        try:
            namespace = '.'.join(self.namespaces)
        except:
            print(self.namespaces)
            raise
        return f':{namespace}.{label}-{self.counter}'

    def assign(self, name, value):
        self.env[name] = value

    def find(self, name, level=0):
        if name in self.env:
            return self.env, level
        elif self.outer:
            return self.outer.find(name, level + 1)
        else:
            return None, None

    def lookup(self, name):
        env, level = self.find(name)
        if env:
            return env[name], level
        else:
            raise RuntimeError(f"undefined variable {name}")

    def child(self, namespace):
        child = Context()
        child.internal = self.internal
        child.outer = self
        child.functions = self.functions
        child.constants = self.constants
        child.namespaces = self.namespaces + [namespace]
        return child


def head(expr):
    if isinstance(expr, Expression) and len(expr) == 0:
        return 'None'
    if isinstance(expr, Expression) and isinstance(expr[0], Symbol):
        return expr[0].name
    if isinstance(expr, Expression):
        return f"OP{expr[0]}"
    if isinstance(expr, Symbol):
        return 'Symbol'
    return 'Atom'


def compile_call(expr, ctx, ret):
    func, *args = expr
    if isinstance(func, Symbol) and func.name == '!asm':
        return list(args)

    code = []
    for arg in reversed(args):
        code.extend(compile_expr(arg, ctx, False))
    code.extend([
        len(args), *compile_expr(func, ctx, ret),
        OPCODE.CALL_RET if ret else OPCODE.CALL
    ])

    return code


def compile_define(expr, ctx, ret):
    is_function = False

    name, value = expr[1:]
    name = name.name
    label = ctx.new_label(name)
    code = compile_expr(value, ctx, False)

    ret = []

    if len(code) == 1 and isinstance(code[0], str):
        if 'lambda' in code[0]:
            is_function = True
            func_code = ctx.functions[code[0]]
            for i in range(len(func_code)):
                if isinstance(func_code[i],
                              str) and func_code[i] == f":external_ref:{name}":
                    func_code[i] = label

    if is_function:
        ctx.functions[label] = ctx.functions.pop(code[0])
    else:
        if len(code) == 2 and code[1] == OPCODE.LOAD:
            label = code[0]
        elif len(code) == 1:
            ctx.constants[label] = value
        else:
            ctx.constants[label] = 0
            ret = [*code, label, OPCODE.STORE]
    ctx.assign(name, label)
    return ret


def compile_lambda(expr, ctx, ret):
    args, body = expr[1:]
    args = [arg.name for arg in args[::-1]]
    label = ctx.new_label('lambda')
    ctx.functions[label] = []
    sub_ctx = ctx.child(label)
    for i, arg in enumerate(args):
        sub_ctx.assign(arg, i)
    code = compile_function_body(body, sub_ctx)
    ctx.functions[label] = code
    return [label]


def compile_function_body(expr, ctx):
    return [*compile_expr(expr, ctx, True), OPCODE.RET]


def compile_symbol(expr, ctx, ret):
    if expr.name in ctx.internal:
        return [ctx.internal[expr.name]]

    try:
        ref, level = ctx.lookup(expr.name)
    except:
        return [f":external_ref:{expr.name}"]

    if ref in ctx.functions:
        return [ref]
    if ref in ctx.constants:
        return [ref, OPCODE.LOAD]

    return [ref, level, OPCODE.SLOAD]


def compile_value(expr, ctx, ret):
    if isinstance(expr, (int, float)):
        return [expr]
    label = ctx.new_label('value')
    ctx.constants[label] = expr
    return [label, OPCODE.LOAD]


def compile_if(expr, ctx, ret):
    cond, then, else_ = expr[1:]
    else_label = ctx.new_label('else')
    end_label = ctx.new_label('end')
    code = compile_expr(cond, ctx, False)
    code.extend([else_label, OPCODE.JE])
    code.extend([*compile_expr(then, ctx, ret), end_label, OPCODE.JMP])
    code.extend([f"label{else_label}"])
    code.extend(compile_expr(else_, ctx, ret))
    code.extend([f"label{end_label}"])
    return code


def compile_cond(expr, ctx, ret):
    # TODO
    code = []
    return code


def compile_begin(expr, ctx, ret):
    code = []
    for e in expr[1:-1]:
        c = compile_expr(e, ctx, False)
        code.extend(c)
        if len(c) > 0:
            code.append(OPCODE.DROP)
    code.extend(compile_expr(expr[-1], ctx, ret))
    return code


def compile_setq(expr, ctx, ret):
    name, value = expr[1:]
    name = name.name
    ref, level = ctx.lookup(name)
    code = compile_expr(value, ctx, False)
    code.extend([OPCODE.DUP, ref, level, OPCODE.SSTORE])
    return code


def compile_let(expr, ctx, ret):
    _, bindings, body = expr
    args = []
    params = []
    for name, value in bindings:
        args.append(name)
        params.append(value)
    expr = Expression(
        [Expression([Symbol('lambda'),
                     Expression(args), body]), *params])
    return compile_expr(expr, ctx, ret)


def compile_let_star(expr, ctx, ret):
    _, bindings, body = expr
    tmp_bindings = []
    inner_bindings = []
    args = []
    for name, _ in bindings:
        args.append(name)
        tmp_bindings.append(Expression([name, 0]))
        inner_bindings.append(Expression([name, name]))
    expr = Expression([
        Symbol('let'),
        Expression([*tmp_bindings]),
        Expression([
            Symbol('begin'), *[
                Expression([Symbol('setq'), name, value])
                for name, value in bindings
            ],
            Expression([Symbol('let'),
                        Expression(inner_bindings), body])
        ])
    ])
    return compile_expr(expr, ctx, ret)


def compile_quote(expr, ctx, ret):
    # TODO
    return [expr]


def compile_while(expr, ctx, ret):
    cond, body = expr[1:]

    loop_start = ctx.new_label('loop')
    loop_end = ctx.new_label('end')
    code = [
        f"label{loop_start}",
        *compile_expr(cond, ctx, False),
        loop_end, OPCODE.JE,
        *compile_expr(body, ctx, ret),
        OPCODE.DROP,
        loop_start, OPCODE.JMP,
        f"label{loop_end}", None
    ] #yapf: disable
    return code


def compile_expr(expr, ctx: Context, ret: bool = False) -> list[Any]:
    dispatch_table = {
        'if': compile_if,
        'cond': compile_cond,
        'begin': compile_begin,
        'setq': compile_setq,
        'let': compile_let,
        'let*': compile_let_star,
        'quote': compile_quote,
        'while': compile_while,
        'lambda': compile_lambda,
        'define': compile_define,
    }

    if isinstance(expr, Expression):
        cond = head(expr)
        if cond == 'None':
            return [None]
        if cond in dispatch_table:
            return dispatch_table[cond](expr, ctx, ret)
        else:
            return compile_call(expr, ctx, ret)
    elif isinstance(expr, Symbol):
        return compile_symbol(expr, ctx, ret)
    else:
        return compile_value(expr, ctx, ret)


def compile(prog: str,
            extra_commands: dict[str, Any] = {},
            language='qlisp') -> list[Any]:
    ctx = Context()
    ctx.internal.update(extra_commands)
    functions = {}

    parsers = {'qlisp': qlisp_parser}
    parse = parsers[language]

    functions['main'] = compile_function_body(parse(prog), ctx)

    for name, code in ctx.functions.items():
        functions[name[1:]] = code

    constants = {}
    for name, value in ctx.constants.items():
        constants[name[1:]] = value

    lib = (functions, constants, ctx.env)

    code = link(functions, constants)
    return code  #, functions, ctx


def _link(code, functions, constants):

    def process(code):
        count = 0
        tmp = []
        labels = {}
        pointers = {}
        functions = {}
        for c in code:
            if isinstance(c, str) and c.startswith('label:'):
                labels[c[6:]] = count
            else:
                count += 1
                tmp.append(c)
                if isinstance(c, str) and c.startswith(':'):
                    pointers[count] = c[1:]
        for i, label in pointers.items():
            if label in labels:
                tmp[i - 1] = labels[label]
            else:
                functions[i - 1] = label
        return tmp, functions

    for name in list(functions.keys()):
        if isinstance(functions[name], tuple):
            continue
        functions[name] = process(functions[name])

    fun_ptrs = {}
    const_ptrs = {}
    code, fun_refs = process(code)

    for name, value in constants.items():
        const_ptrs[name] = len(code)
        code.append(value)

    for name, (c, f_refs) in functions.items():
        fun_ptrs[name] = len(code)
        code.extend(c)
        for i, n in f_refs.items():
            fun_refs[i + fun_ptrs[name]] = n
    external_refs = {}
    for addr, label in fun_refs.items():
        if label in fun_ptrs:
            code[addr] = fun_ptrs[label]
        elif label in const_ptrs:
            code[addr] = const_ptrs[label]
        else:
            external_refs[addr] = label
    return code, external_refs


def link(functions=None, constants=None, dynamic=False):
    if functions is None:
        functions = {}
    if constants is None:
        constants = {}
    if "main" not in functions and not dynamic:
        raise RuntimeError("main function not defined")

    enter_code = [0, ":main", OPCODE.CALL, OPCODE.EXIT]
    code, external_refs = _link(enter_code, functions, constants)
    if external_refs and not dynamic:
        raise RuntimeError(
            f"external references: {set(external_refs.values())}")
    return code



import itertools
from functools import partial, reduce

import numpy as np

from .mat import U, fSim, make_immutable, rfUnitary

try:
    import tensornetwork as tn
except ImportError:
    pass

__matrix_of_gates = {}


def regesterGateMatrix(gate, mat, N=None):
    if isinstance(mat, np.ndarray):
        mat = make_immutable(mat)
    if N is None:
        N = int(np.log2(mat.shape[0]))
    __matrix_of_gates[gate] = (mat, N)


def gate2mat(gate):
    if isinstance(gate, str) and gate in __matrix_of_gates:
        return __matrix_of_gates[gate][0]
    elif isinstance(gate, tuple) and gate[0] in __matrix_of_gates:
        if callable(__matrix_of_gates[gate[0]][0]):
            return __matrix_of_gates[gate[0]][0](*gate[1:])
        else:
            raise ValueError(
                f"Could not call {gate[0]}(*{gate[1:]}), `{gate[0]}` is not callable."
            )
    elif isinstance(gate, tuple) and gate[0] == 'C':
        U = gate2mat(gate[1])
        ret = np.eye(2 * U.shape[0], dtype=np.complex)
        ret[U.shape[0]:, U.shape[0]:] = U
        return ret
    else:
        raise ValueError(f'Unexcept gate {gate}')


def splite_at(l, bits):
    """将 l 的二进制位于 bits 所列的位置上断开插上0
    
    如 splite_at(int('1111',2), [0,2,4,6]) == int('10101010', 2)
    bits 必须从小到大排列
    """
    r = l
    for n in bits:
        mask = (1 << n) - 1
        low = r & mask
        high = r - low
        r = (high << 1) + low
    return r


def place_at(l, bits):
    """将 l 的二进制位置于 bits 所列的位置上
    
    如 place_at(int('10111',2), [0,2,4,5,6]) == int('01010101', 2)
    """
    r = 0
    for index, n in enumerate(bits):
        b = (l >> index) & 1
        r += b << n
    return r


def reduceSubspace(targets, N, inputMat, func, args):
    innerDim = 2**len(targets)
    outerDim = 2**(N - len(targets))

    targets = tuple(reversed([N - i - 1 for i in targets]))

    def index(targets, i, j):
        return splite_at(j, sorted(targets)) | place_at(i, targets)

    if len(inputMat.shape) == 1:
        for k in range(outerDim):
            innerIndex = [index(targets, i, k) for i in range(innerDim)]
            inputMat[innerIndex] = func(inputMat[innerIndex], *args)
    else:
        for k, l in itertools.product(range(outerDim), repeat=2):
            innerIndex = np.asarray(
                [[index(targets, i, k),
                  index(targets, j, l)]
                 for i, j in itertools.product(range(innerDim), repeat=2)]).T
            sub = inputMat[innerIndex[0], innerIndex[1]].reshape(
                (innerDim, innerDim))
            inputMat[innerIndex[0], innerIndex[1]] = func(sub, *args).flatten()
    return inputMat


def applySeq(seq, psi0=None):
    if psi0 is None:
        psi = np.array([1, 0], dtype=complex)
        N = 1
    else:
        psi = psi0
        N = int(np.round(np.log2(psi.shape[0])))

    if len(psi.shape) == 1:
        I = np.array([1, 0])
    else:
        I = np.eye(2)

    def func(psi, U):
        return U @ psi

    for gate, qubits in seq:
        if (isinstance(gate, tuple) and gate[0] in ['Measure', 'Delay']) or (
                isinstance(gate, str) and gate in ['Barrier']):
            continue
        if isinstance(qubits, tuple):
            M = max(qubits)
        else:
            M = qubits
            qubits = (qubits, )
        if M >= N:
            psi = reduce(np.kron, itertools.repeat(I, times=M - N + 1), psi)
            N = M + 1

        reduceSubspace(qubits, N, psi, func, (gate2mat(gate), ))

    return psi


def seq2mat(seq):
    return applySeq(seq, np.eye(2, dtype=complex))


def apply_gate(qubit_edges, gate, operating_qubits):
    if 'tn' not in globals():
        raise ImportError('Please install tensornetwork first.')

    op = tn.Node(gate)
    for i, bit in enumerate(operating_qubits):
        tn.connect(qubit_edges[bit], op[i])
        qubit_edges[bit] = op[i + len(operating_qubits)]


def circuit_network(circ):
    if 'tn' not in globals():
        raise ImportError('Please install tensornetwork first.')

    N = 0

    all_nodes = []

    with tn.NodeCollection(all_nodes):
        left_edges = []
        right_edges = []

        for gate, qubits in circ:
            if (isinstance(gate, tuple)
                    and gate[0] in ['Measure', 'Delay']) or (isinstance(
                        gate, str) and gate in ['Barrier']):
                continue
            if isinstance(qubits, tuple):
                M = max(qubits)
            else:
                M = qubits
                qubits = (qubits, )
            if M >= N:
                new_nodes = [tn.Node(np.eye(2)) for _ in range(M - N + 1)]
                left_edges.extend([node[0] for node in new_nodes])
                right_edges.extend([node[1] for node in new_nodes])
                N = M + 1

            gate_mat = gate2mat(gate)
            gate_tenser = gate_mat.reshape((2, 2) * len(qubits))
            apply_gate(left_edges, gate_tenser, qubits)

    return all_nodes, left_edges, right_edges, N


def apply_circuit(circ, qubits=None, init_state=None):
    if 'tn' not in globals():
        raise ImportError('Please install tensornetwork first.')

    all_nodes, left_edges, right_edges, N = circuit_network(circ)
    if qubits is not None:
        assert len(qubits) == N
    else:
        qubits = list(range(N))
    if init_state is None:
        init_state_nodes = [
            tn.Node(np.array([1, 0], dtype=complex)) for i in range(N)
        ]
        init_state_edges = [init_state_nodes[i][0] for i in range(N)]
    else:
        init_state_nodes, init_state_edges = init_state

    all_nodes.extend(init_state_nodes)
    qubit_edges = [left_edges[q] for q in qubits]
    for i, q in enumerate(qubits):
        tn.connect(right_edges[i], init_state_edges[q])
    result = tn.contractors.optimal(all_nodes, output_edge_order=qubit_edges)
    return result.tensor.reshape(-1)


def circuit2mat(circ):
    if 'tn' not in globals():
        raise ImportError('Please install tensornetwork first.')

    all_nodes, left_edges, right_edges, N = circuit_network(circ)
    result = tn.contractors.optimal(all_nodes,
                                    output_edge_order=left_edges + right_edges)
    return result.tensor.reshape((2**N, 2**N))


regesterGateMatrix('U', U, 1)
regesterGateMatrix('P', lambda p: U(theta=0, phi=0, lambda_=p), 1)
regesterGateMatrix('rfUnitary', rfUnitary, 1)
regesterGateMatrix('Rx', partial(rfUnitary, phi=0), 1)
regesterGateMatrix('Ry', partial(rfUnitary, phi=np.pi / 2), 1)
regesterGateMatrix('Rz', lambda p: U(theta=0, phi=0, lambda_=p), 1)
regesterGateMatrix('fSim', fSim, 2)

# one qubit
regesterGateMatrix('I', np.array([[1, 0], [0, 1]]))
regesterGateMatrix('X', np.array([[0, -1j], [-1j, 0]]))
regesterGateMatrix('Y', np.array([[0, -1], [1, 0]]))
regesterGateMatrix('X/2', np.array([[1, -1j], [-1j, 1]]) / np.sqrt(2))
regesterGateMatrix('Y/2', np.array([[1, -1], [1, 1]]) / np.sqrt(2))
regesterGateMatrix('-X/2', np.array([[1, 1j], [1j, 1]]) / np.sqrt(2))
regesterGateMatrix('-Y/2', np.array([[1, 1], [-1, 1]]) / np.sqrt(2))
regesterGateMatrix('Z', np.array([[1, 0], [0, -1]]))
regesterGateMatrix('S', np.array([[1, 0], [0, 1j]]))
regesterGateMatrix('-S', np.array([[1, 0], [0, -1j]]))
regesterGateMatrix('H', np.array([[1, 1], [1, -1]]) / np.sqrt(2))

# non-clifford
regesterGateMatrix('T',
                   np.array([[1, 0], [0, 1 / np.sqrt(2) + 1j / np.sqrt(2)]]))
regesterGateMatrix('-T',
                   np.array([[1, 0], [0, 1 / np.sqrt(2) - 1j / np.sqrt(2)]]))
regesterGateMatrix('W/2', rfUnitary(np.pi / 2, np.pi / 4))
regesterGateMatrix('-W/2', rfUnitary(-np.pi / 2, np.pi / 4))
regesterGateMatrix('V/2', rfUnitary(np.pi / 2, 3 * np.pi / 4))
regesterGateMatrix('-V/2', rfUnitary(-np.pi / 2, 3 * np.pi / 4))

# two qubits
regesterGateMatrix(
    'CZ', np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]]))
regesterGateMatrix(
    'Cnot', np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]))
regesterGateMatrix(
    'iSWAP',
    np.array([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]]))
regesterGateMatrix(
    'SWAP', np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]))
regesterGateMatrix(
    'CR',
    np.array([[1, 1j, 0, 0], [1j, 1, 0, 0], [0, 0, 1, -1j], [0, 0, -1j, 1]]) /
    np.sqrt(2))

# non-clifford
regesterGateMatrix(
    'SQiSWAP',
    np.array([[1, 0, 0, 0], [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0],
              [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0], [0, 0, 0, 1]]))

if __name__ == '__main__':
    # Porter-Thomas distribution

    def randomSeq(depth, N):
        seq = []
        for i in range(depth):
            for j in range(N):
                seq.append((np.random.choice(['X/2', 'Y/2', 'W/2']), j))
            for j in range(i % 2, N, 2):
                seq.append(('SQiSWAP', (j, (j + 1) % N)))
        return seq

    p = []
    # run 1000 random circuit on 6 qubits
    for i in range(1000):
        print('    ', i, end='')
        seq = randomSeq(50, 6)
        psi = applySeq(seq)
        p.extend(list(np.abs(psi)**2))
        print('    ', i)
    p = np.asarray(p)

    # plot distribution of probabilities
    N = 2**6
    y, x = np.histogram(N * p, bins=50, density=True)

    import matplotlib.pyplot as plt

    plt.semilogy((x[:-1] + x[1:]) / 2, y)
    plt.show()



import numpy as np


def fit_k(x, y, e=None, axis=None):
    """
    Fit a line to data.
    y = k * x
    """
    if e is None:
        e = np.full_like(y, 1.0)
    weight = 1 / e**2

    return np.sum(x * y * weight, axis=axis) / np.sum(x**2 * weight, axis=axis)


def lin_fit(x, y, axis=None):
    """use less memory than np.polyfit"""
    x, y = np.asarray(x), np.asarray(y)
    xm, ym = x.mean(axis=axis), y.mean(axis=axis)
    N = len(x)
    a = (np.sum(x * y, axis=axis) - N * xm * ym) / (
        (x**2).sum(axis=axis) - N * xm * xm)
    b = ym - a * xm
    return np.array([a, b])


def poly_fit(x_data, y_data, degree=20, errors=None, selectors=[]):
    """
    Fit a polynomial to data.

    Parameters
    ----------
    x_data : array_like
        x data.
    y_data : array_like
        y data.
    degree : int, optional
        Degree of polynomial. The default is 20.
    errors : float, optional
        errors of y_data. The default is None.
    selectors : list, optional
        List of pairs of weights. Repeatedly fit the data
        with the selected data.

        err = (y_fit - y_data)**2
        thr = np.mean(err) * a + np.median(err) * b
        selector = err < thr
    
    Returns
    -------
    a : array_like
        Coefficients of polynomial.
    lim : list
        Limit of x data.
    mask : array_like
        Mask of selected data.

    Examples
    --------
    >>> import numpy as np
    >>> from waveforms.math.fit import poly_fit
    >>> x = np.linspace(0, 10, 30)
    >>> f = np.poly1d([0.3, 0, 0, 0, 0, 0, 2, 1, 0])
    >>> y = f(x) + np.random.normal(0, 0.1, 20)
    >>> y[4] += 3
    >>> y[16] -= 2
    >>> a, lim, mask = poly_fit(x, y, 20, [(0.1, 1), (3, 0.8), (3, 0.9)])
    >>> mask
    array([ True,  True,  True,  True, False,  True,  True,  True,  True,
            True,  True,  True,  True,  True,  True,  True, False,  True,
            True,  True])
    >>> lim
    [0.0, 1.0]
    """
    try:
        from sklearn.linear_model import Ridge
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import PolynomialFeatures, StandardScaler
    except ImportError:
        raise ImportError('scikit-learn is required for poly_fit')

    poly = PolynomialFeatures(degree=degree)
    poly.fit(x_data.reshape((-1, 1)))
    x_train = poly.transform(x_data.reshape((-1, 1)))

    model = Pipeline([
        ('sca', StandardScaler()),
        ('ridge', Ridge(solver='cholesky')),
    ])

    if errors is None:
        weight = np.ones_like(y_data)
    else:
        weight = 1 / errors**2

    model.fit(x_train, y_data, sample_weight=weight)
    y_fit = model.predict(x_train)

    # select data
    for a, b in selectors:
        err = (y_fit - y_data)**2
        thr = (np.mean(err) * a + np.median(err) * b) / weight
        mask = err < thr
        model.fit(x_train[mask],
                  np.array(y_data)[mask],
                  sample_weight=weight[mask])
        y_fit = model.predict(x_train)

    x, y = x_data[mask].reshape(-1), np.array(y_data)[mask]
    xx = np.linspace(x[0], x[-1], degree * 10)
    poly = PolynomialFeatures(degree=degree)
    poly.fit(xx.reshape((-1, 1)))
    xxx = poly.transform(xx.reshape((-1, 1)))
    a = np.polyfit(xx, model.predict(xxx), degree)

    return a, [np.min(x_data), np.max(x_data)], mask


def inv_poly(x, a, lim=None):
    """
    Find the root of polynomial.

    Parameters
    ----------
    x : float
        Value of polynomial.
    a : array_like
        Coefficients of polynomial.
    lim : list, optional
        Limit of root. The default is None.

    Returns
    -------
    float
        Root of polynomial.

    Raises
    ------
    ValueError
        Can not find root in range lim.
    """
    a = list(a)
    a[-1] = a[-1] - x
    roots = np.poly1d(a).roots
    y = roots[np.abs(roots.imag) <= 1e-12].real
    if lim is None:
        return y
    for b in y:
        if lim[0] <= b <= lim[1]:
            return b
    raise ValueError(f'Can not find root in range {lim}.')


def fit_circle(x, y):
    """
    Fit a circle to data.
    """
    u, v = x - x.mean(), y - y.mean()
    Suuu, Svvv = np.sum(u**3), np.sum(v**3)
    Suu, Svv = np.sum(u**2), np.sum(v**2)
    Suv = np.sum(u * v)
    Suuv, Suvv = np.sum(u**2 * v), np.sum(u * v**2)
    uc = (Suuv * Suv - Suuu * Svv - Suvv * Svv + Suv * Svvv)
    vc = (-Suu * Suuv + Suuu * Suv + Suv * Suvv - Suu * Svvv)
    uc /= 2 * (Suv**2 - Suu * Svv)
    vc /= 2 * (Suv**2 - Suu * Svv)
    xc, yc = uc + x.mean(), vc + y.mean()
    R = np.sqrt(np.mean((x - xc)**2 + (y - yc)**2))
    return xc, yc, R


def fit_cross_point(x1, y1, x2, y2):
    a1, b1 = lin_fit(x1, y1)
    a2, b2 = lin_fit(x2, y2)
    return (b2 - b1) / (a1 - a2), (a1 * b2 - a2 * b1) / (a1 - a2)


def find_cross_point(z1, z2, z3, z4):
    v1 = z2 - z1
    v2 = z4 - z3
    v3 = z3 - z1
    a = (v1.real * v2.imag - v1.imag * v2.real)
    t1 = (v3.real * v2.imag - v3.imag * v2.real) / a
    t2 = (v3.real * v1.imag - v3.imag * v1.real) / a

    return z1 + t1 * v1, t1, t2


def fit_pole(x, y):
    a, b, c = np.polyfit(x, y, 2)
    return -0.5 * b / a, c - 0.25 * b**2 / a


def fit_cosine(data, repeat=1, weight=None, x=None):
    """
    Find the amplitude and phase of the data.
    solve equations:
        data[i] = offset + R * cos(phi + i * repeat * 2 * pi / N)
        where i = 0, 1, ..., N-1 and N = len(data)

    Args:
        data (array): data
        repeat (int, optional): Number of cycles. Defaults to 1.
        weight (array, optional): weight. Defaults to None.
    
    Returns:
        tuple: (R, offset, phi)
    
    Examples:
        >>> import numpy as np
        >>> np.random.seed(1234)
        >>> N = 100
        >>> x = 2 * np.pi * np.linspace(0, 1, N, endpoint=False)
        >>> y = 1.2 * np.cos(x - 0.9) + 0.4 + 0.01*np.random.randn(N)
        >>> R, offset, phi = fit_cosine(y)
        >>> R, offset, phi
        (1.20002782555664, 0.4003511228312543, -0.9002458177749354)
    """
    data = np.asarray(data)
    N = data.shape[0]
    if x is None:
        x = np.linspace(0, 2 * np.pi * repeat, N, endpoint=False)
    else:
        weight = np.ones_like(x)
    if weight is None:
        offset = data.mean(axis=0)
        data = data - offset
        e = np.sum(np.moveaxis(data, 0, -1) * np.cos(x), axis=-1)
        f = np.sum(np.moveaxis(data, 0, -1) * np.sin(x), axis=-1)
        R = 2 * np.sqrt(e**2 + f**2) / N
        phi = np.arctan2(-f, e)
    else:
        if weight.ndim == 1:
            assert weight.shape[0] == N, "len(weight) != data.shape[0]"
        elif weight.ndim == 2:
            assert weight.shape == data.shape, "weight.shape != data.shape"
        else:
            raise ValueError("weight must be 1d or 2d array")
        weight = weight / weight.sum(axis=0)
        y = np.sum(np.moveaxis(data, 0, -1) * np.moveaxis(weight, 0, -1),
                   axis=-1)

        a = np.sum(np.moveaxis(weight, 0, -1) * np.cos(x), axis=-1)
        b = np.sum(np.moveaxis(weight, 0, -1) * np.sin(x), axis=-1)
        c = np.sum(np.moveaxis(weight, 0, -1) * np.cos(2 * x), axis=-1)
        d = np.sum(np.moveaxis(weight, 0, -1) * np.sin(2 * x), axis=-1)
        e = np.sum(np.moveaxis(data, 0, -1) * np.moveaxis(weight, 0, -1) *
                   np.cos(x),
                   axis=-1)
        f = np.sum(np.moveaxis(data, 0, -1) * np.moveaxis(weight, 0, -1) *
                   np.sin(x),
                   axis=-1)
        offset = (2 * a * c * e - 2 * a * e + 2 * b * d * e - 2 * b * f -
                  2 * b * c * f + 2 * a * d * f + y - c**2 * y -
                  d**2 * y) / (1 - 2 * a**2 - 2 * b**2 + 2 * a**2 * c -
                               2 * b**2 * c - c**2 + 4 * a * b * d - d**2)
        data = data - offset
        e = np.sum(np.moveaxis(data, 0, -1) * np.moveaxis(weight, 0, -1) *
                   np.cos(x),
                   axis=-1)
        f = np.sum(np.moveaxis(data, 0, -1) * np.moveaxis(weight, 0, -1) *
                   np.sin(x),
                   axis=-1)
        R = (2 * np.sqrt(((c - 1)**2 + d**2) * e**2 + (
            (c + 1)**2 + d**2) * f**2 - 4 * d * e * f)) / (c**2 + d**2 - 1)
        phi = np.arctan2(e - c * e - d * f, f + c * f - d * e) + np.pi / 2

    return R, offset, phi


def complex_amp_to_real(s, axis=None):
    k, _ = lin_fit(s.real, s.imag, axis=axis)
    phi = np.arctan(k)
    return np.real(s * np.exp(-1j * phi))


def goodness_of_fit(pnum, ydata, fvec, sigma=None):
    """
    拟合优度

    Args:
        pnum (int): 拟合参数的个数
        ydata (1d-array): 原始数据
        fvec (1d-array): 残差
        sigma (1d-array, optional): 标准差. Defaults to None.

    Returns:
        tuple: (SSE, R_square, Adj_R_sq, RMSE)
    """
    n = len(ydata)
    dfe = n - pnum
    SSE = (fvec**2).sum()
    if sigma is not None:
        SST = (((ydata - ydata.mean()) / sigma)**2).sum()
    else:
        std = np.sqrt(1.0 * n / dfe * (fvec**2).mean())
        SST = (((ydata - ydata.mean()) / std)**2).sum()
    RMSE = np.sqrt(SSE / dfe)
    R_square = 1 - SSE / SST
    Adj_R_sq = 1 - (SSE / (dfe - 1)) / (SST / (n - 1))

    return SSE, R_square, Adj_R_sq, RMSE



from concurrent.futures import process
import itertools
from functools import partial, reduce

import numpy as np

from .mat import U, fSim, make_immutable, rfUnitary

__matrix_of_gates = {}


def regesterGateMatrix(gate, mat, N=None, docs=''):
    if isinstance(mat, np.ndarray):
        mat = make_immutable(mat)
    if N is None:
        N = round(np.log2(mat.shape[0]))
    __matrix_of_gates[gate] = (mat, N, docs)


def gate_name(gate):
    if isinstance(gate, tuple):
        return gate_name(gate[0])
    elif isinstance(gate, str):
        return gate
    else:
        raise ValueError(f'Unexcept gate {gate}')


def gate2mat(gate):
    if isinstance(gate, str) and gate in __matrix_of_gates:
        return __matrix_of_gates[gate][:2]
    elif isinstance(gate, tuple) and gate[0] in __matrix_of_gates:
        if callable(__matrix_of_gates[gate[0]][0]):
            return __matrix_of_gates[gate[0]][0](
                *gate[1:]), __matrix_of_gates[gate[0]][1]
        else:
            raise ValueError(
                f"Could not call {gate[0]}(*{gate[1:]}), `{gate[0]}` is not callable."
            )
    elif gate_name(gate) == 'C':
        U, N = gate2mat(gate[1])
        ret = np.eye(2 * U.shape[0], dtype=np.complex)
        ret[U.shape[0]:, U.shape[0]:] = U
        return ret, N + 1
    else:
        raise ValueError(f'Unexcept gate {gate}')


def splite_at(l, bits):
    """将 l 的二进制位于 bits 所列的位置上断开插上0
    
    如 splite_at(int('1111',2), [0,2,4,6]) == int('10101010', 2)
    bits 必须从小到大排列
    """
    r = l
    for n in bits:
        mask = (1 << n) - 1
        low = r & mask
        high = r - low
        r = (high << 1) + low
    return r


def place_at(l, bits):
    """将 l 的二进制位置于 bits 所列的位置上
    
    如 place_at(int('10111',2), [0,2,4,5,6]) == int('01010101', 2)
    """
    r = 0
    for index, n in enumerate(bits):
        b = (l >> index) & 1
        r += b << n
    return r


def reduceSubspace(targets, N, inputMat, func, args):
    innerDim = 2**len(targets)
    outerDim = 2**(N - len(targets))

    targets = tuple(reversed([N - i - 1 for i in targets]))

    def index(targets, i, j):
        return splite_at(j, sorted(targets)) | place_at(i, targets)

    if len(inputMat.shape) == 1:
        for k in range(outerDim):
            innerIndex = [index(targets, i, k) for i in range(innerDim)]
            inputMat[innerIndex] = func(inputMat[innerIndex], *args)
    else:
        for k, l in itertools.product(range(outerDim), repeat=2):
            innerIndex = np.asarray(
                [[index(targets, i, k),
                  index(targets, j, l)]
                 for i, j in itertools.product(range(innerDim), repeat=2)]).T
            sub = inputMat[innerIndex[0], innerIndex[1]].reshape(
                (innerDim, innerDim))
            inputMat[innerIndex[0], innerIndex[1]] = func(sub, *args).flatten()
    return inputMat


def _apply_gate(gate, inputMat, unitary_process, qubits, N):
    U, n = gate2mat(gate)
    if len(qubits) == n and all(isinstance(qubit, int) for qubit in qubits):
        reduceSubspace(qubits, N, inputMat, unitary_process, (U, ))
    elif n == 1 and all(isinstance(qubit, int) for qubit in qubits):
        for qubit in qubits:
            reduceSubspace([qubit], N, inputMat, unitary_process, (U, ))
    elif len(qubits) == n and all(
            isinstance(qubit, [tuple, list]) for qubit in qubits):
        for qubit_tuple in zip(*qubits):
            reduceSubspace(qubit_tuple, N, inputMat, unitary_process, (U, ))
    else:
        raise ValueError(f'Unexcept gate {gate} and qubits {qubits}')


def _measure_process(rho):
    return np.array([[rho[0, 0], 0], [0, rho[1, 1]]])


def _reset_process(rho):
    return np.array([[rho[0, 0] + rho[1, 1], 0], [0, 0]])


def _decohherence_process(rho, Gamma_t, gamma_t):
    rho00 = rho[0, 0] + rho[1, 1] * (1 - np.exp(-Gamma_t))
    rho11 = rho[1, 1] * np.exp(-Gamma_t)
    rho01 = rho[0, 1] * np.exp(-Gamma_t / 2 - gamma_t**2)
    rho10 = rho[1, 0] * np.exp(-Gamma_t / 2 - gamma_t**2)
    return np.array([[rho00, rho01], [rho10, rho11]])


def applySeq(seq, psi0=None):

    def _set_vector_to_rho(psi):
        psi = psi.reshape(-1, 1).conj() @ psi.reshape(1, -1)
        unitary_process = lambda psi, U: U @ psi @ U.T.conj()
        return psi, unitary_process, np.array([[1, 0], [0, 0]])

    if psi0 is None:
        psi = np.array([1, 0], dtype=complex)
        N = 1
    else:
        psi = psi0
        N = round(np.log2(psi.shape[0]))

    psi0 = np.array([1, 0])
    unitary_process = lambda psi, U: U @ psi

    if psi.ndim == 2:
        psi, unitary_process, psi0 = _set_vector_to_rho(psi)

    for gate, *qubits in seq:
        M = max(qubits)
        if M >= N:
            psi = reduce(np.kron, itertools.repeat(psi0, times=M - N + 1), psi)
            N = M + 1

        if gate_name(gate) in ['Barrier']:
            continue
        if gate_name(gate) in ['Delay']:
            if len(gate) == 2:
                continue
            else:
                if len(gate) == 3:
                    _, t, T1 = gate
                    Gamma_t = t / T1
                    gamma_t = 0
                else:
                    _, t, T1, Tphi = gate
                    Gamma_t = t / T1
                    gamma_t = t / Tphi
        if gate_name(gate) in ['Measure', 'Reset', 'Delay'] and psi.ndim == 1:
            psi, unitary_process, psi0 = _set_vector_to_rho(psi)

        if gate_name(gate) == 'Measure':
            reduceSubspace(qubits, N, psi, _measure_process, ())
        elif gate_name(gate) == 'Reset':
            reduceSubspace(qubits, N, psi, _reset_process, ())
        elif gate_name(gate) == 'Delay':
            reduceSubspace(qubits, N, psi, _decohherence_process,
                           (Gamma_t, gamma_t))
        else:
            _apply_gate(gate, psi, unitary_process, qubits, N)

    return psi


def seq2mat(seq, U=None):
    I = np.eye(2, dtype=np.complex)
    if U is None:
        U = np.eye(2, dtype=np.complex)
        N = 1
    else:
        N = round(np.log2(U.shape[0]))

    unitary_process = lambda U0, U: U @ U0

    for gate, *qubits in seq:
        M = max(qubits)
        if M >= N:
            U = reduce(np.kron, itertools.repeat(I, times=M - N + 1), U)
            N = M + 1
        if gate_name(gate) in ['Delay', 'Barrier']:
            continue
        if gate_name(gate) in ['Measure', 'Reset']:
            raise ValueError(
                'Measure and Reset must be applied to a state vector')
        else:
            _apply_gate(gate, U, unitary_process, qubits, N)
    return U


regesterGateMatrix('U', U, 1)
regesterGateMatrix('P', lambda p: U(theta=0, phi=0, lambda_=p), 1)
regesterGateMatrix('rfUnitary', rfUnitary, 1)
regesterGateMatrix('Rx', partial(rfUnitary, phi=0), 1)
regesterGateMatrix('Ry', partial(rfUnitary, phi=np.pi / 2), 1)
regesterGateMatrix('Rz', lambda p: U(theta=0, phi=0, lambda_=p), 1)
regesterGateMatrix('fSim', fSim, 2)

# one qubit
regesterGateMatrix('I', np.array([[1, 0], [0, 1]]))
regesterGateMatrix('X', np.array([[0, -1j], [-1j, 0]]))
regesterGateMatrix('Y', np.array([[0, -1], [1, 0]]))
regesterGateMatrix('X/2', np.array([[1, -1j], [-1j, 1]]) / np.sqrt(2))
regesterGateMatrix('Y/2', np.array([[1, -1], [1, 1]]) / np.sqrt(2))
regesterGateMatrix('-X/2', np.array([[1, 1j], [1j, 1]]) / np.sqrt(2))
regesterGateMatrix('-Y/2', np.array([[1, 1], [-1, 1]]) / np.sqrt(2))
regesterGateMatrix('Z', np.array([[1, 0], [0, -1]]))
regesterGateMatrix('S', np.array([[1, 0], [0, 1j]]))
regesterGateMatrix('-S', np.array([[1, 0], [0, -1j]]))
regesterGateMatrix('H', np.array([[1, 1], [1, -1]]) / np.sqrt(2))

# non-clifford
regesterGateMatrix('T',
                   np.array([[1, 0], [0, 1 / np.sqrt(2) + 1j / np.sqrt(2)]]))
regesterGateMatrix('-T',
                   np.array([[1, 0], [0, 1 / np.sqrt(2) - 1j / np.sqrt(2)]]))
regesterGateMatrix('W/2', rfUnitary(np.pi / 2, np.pi / 4))
regesterGateMatrix('-W/2', rfUnitary(-np.pi / 2, np.pi / 4))
regesterGateMatrix('V/2', rfUnitary(np.pi / 2, 3 * np.pi / 4))
regesterGateMatrix('-V/2', rfUnitary(-np.pi / 2, 3 * np.pi / 4))

# two qubits
regesterGateMatrix(
    'CZ', np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]]))
regesterGateMatrix(
    'Cnot', np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]))
regesterGateMatrix(
    'iSWAP',
    np.array([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]]))
regesterGateMatrix(
    'SWAP', np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]))
regesterGateMatrix(
    'CR',
    np.array([[1, 1j, 0, 0], [1j, 1, 0, 0], [0, 0, 1, -1j], [0, 0, -1j, 1]]) /
    np.sqrt(2))

# non-clifford
regesterGateMatrix(
    'SQiSWAP',
    np.array([[1, 0, 0, 0], [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0],
              [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0], [0, 0, 0, 1]]))

if __name__ == '__main__':
    # Porter-Thomas distribution

    def randomSeq(depth, N):
        seq = []
        for i in range(depth):
            for j in range(N):
                seq.append((np.random.choice(['X/2', 'Y/2', 'W/2']), j))
            for j in range(i % 2, N, 2):
                seq.append(('SQiSWAP', j, (j + 1) % N))
        return seq

    p = []
    # run 1000 random circuit on 6 qubits
    for i in range(1000):
        print('    ', i, end='')
        seq = randomSeq(50, 6)
        psi = applySeq(seq)
        p.extend(list(np.abs(psi)**2))
        print('    ', i)
    p = np.asarray(p)

    # plot distribution of probabilities
    N = 2**6
    y, x = np.histogram(N * p, bins=50, density=True)

    import matplotlib.pyplot as plt

    plt.semilogy((x[:-1] + x[1:]) / 2, y)
    plt.show()



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Optional

from waveforms.quantum.circuit.qlisp.arch.base import (COMMAND, DataMap,
                                                       QLispCode)
from waveforms.quantum.circuit.qlisp.qlisp import Signal
from waveforms.scan_iter import StepStatus


@dataclass
class ProgramFrame():
    """
    A frame of a program.
    """
    step: StepStatus = field(default=None)
    cmds: list[COMMAND] = field(default_factory=list)
    data_map: DataMap = field(default_factory=dict)
    code: Optional[QLispCode] = None

    fut: asyncio.Future = None

    def __getstate__(self):
        state = self.__dict__.copy()
        try:
            del state['fut']
        except:
            pass
        return state


@dataclass
class Program:
    """
    A program is a list of commands.
    """
    with_feedback: bool = False
    arch: str = 'baqis'

    side_effects: dict = field(default_factory=dict)

    steps: list[ProgramFrame] = field(default_factory=list)
    shots: int = 1024
    signal: Signal = Signal.state

    snapshot: dict = field(default_factory=dict)
    task_arguments: tuple[tuple, dict] = (tuple(), dict())
    meta_info: dict = field(default_factory=dict)



import pickle
from bisect import bisect_left
from itertools import chain, product
from math import comb

import numpy as np
import scipy.special as special
from numpy import e, inf, pi

NDIGITS = 15

_zero = ((), ())


def _const(c):
    if c == 0:
        return _zero
    return (((), ()), ), (c, )


_one = _const(1.0)
_half = _const(1 / 2)
_two = _const(2.0)
_pi = _const(pi)
_two_pi = _const(2 * pi)
_half_pi = _const(pi / 2)


def _is_const(x):
    return x == _zero or x[0] == (((), ()), )


def _basic_wave(Type, *args, shift=0):
    return ((((Type, *args, shift), ), (1, )), ), (1.0, )


def _insert_type_value_pair(t_list, v_list, t, v, lo, hi):
    i = bisect_left(t_list, t, lo, hi)
    if i < hi and t_list[i] == t:
        v += v_list[i]
        if v == 0:
            t_list.pop(i)
            v_list.pop(i)
            return i, hi - 1
        else:
            v_list[i] = v
            return i, hi
    else:
        t_list.insert(i, t)
        v_list.insert(i, v)
        return i, hi + 1


def _mul(x, y):
    t_list, v_list = [], []
    xt_list, xv_list = x
    yt_list, yv_list = y
    lo, hi = 0, 0
    for (t1, t2), (v1, v2) in zip(product(xt_list, yt_list),
                                  product(xv_list, yv_list)):
        if v1 * v2 == 0:
            continue
        t = _add(t1, t2)
        lo, hi = _insert_type_value_pair(t_list, v_list, t, v1 * v2, lo, hi)
    return tuple(t_list), tuple(v_list)


def _add(x, y):
    #x, y = (x, y) if len(x[0]) >= len(y[0]) else (y, x)
    t_list, v_list = list(x[0]), list(x[1])
    lo, hi = 0, len(t_list)
    for t, v in zip(*y):
        lo, hi = _insert_type_value_pair(t_list, v_list, t, v, lo, hi)
    return tuple(t_list), tuple(v_list)


def _shift(x, time):
    if _is_const(x):
        return x

    t_list = []

    for pre_mtlist, nlist in x[0]:
        mtlist = []
        for Type, *args, shift in pre_mtlist:
            mtlist.append((Type, *args, shift + time))
        t_list.append((tuple(mtlist), nlist))
    return tuple(t_list), x[1]


def _pow(x, n):
    if x == _zero:
        return _zero
    if n == 0:
        return _one
    if _is_const(x):
        return _const(x[1][0]**n)

    if len(x[0]) == 1:
        t_list, v_list = [], []
        for (mtlist, pre_nlist), v in zip(*x):
            nlist = []
            for m in pre_nlist:
                nlist.append(n * m)
            t_list.append((mtlist, tuple(nlist)))
            v_list.append(v**n)
        return tuple(t_list), tuple(v_list)
    else:
        assert isinstance(n, int) and n > 0
        ret = _one
        for i in range(n):
            ret = _mul(ret, x)
        return ret


def _cos_power_n(x, n):
    _, w, shift = x
    ret = _zero
    for k in range(0, n // 2 + 1):
        if n == 2 * k:
            a = _const(comb(n, k) / 2**n)
            ret = _add(ret, a)
        else:
            expr = (((((COS, (n - 2 * k) * w, shift), ), (1, )), ),
                    (comb(n, k) / 2**(n - 1), ))
            ret = _add(ret, expr)
    return ret


def _trigMul_t(x, y, v):
    """cos(a)cos(b) = 0.5*cos(a+b)+0.5*cos(a-b)"""
    _, w1, t1 = x
    _, w2, t2 = y
    if w2 > w1:
        t1, t2 = t2, t1
        w1, w2 = w2, w1
    exp1 = (COS, w1 + w2, (w1 * t1 + w2 * t2) / (w1 + w2))
    if w1 == w2:
        c = v * np.cos(w1 * t1 - w2 * t2) / 2
        if c == 0:
            return (((exp1, ), (1, )), ), (0.5 * v, )
        else:
            return (((), ()), ((exp1, ), (1, ))), (c, 0.5 * v)
    else:
        exp2 = (COS, w1 - w2, (w1 * t1 - w2 * t2) / (w1 - w2))
        if exp2[1] > exp1[1]:
            exp2, exp1 = exp1, exp2
        return (((exp2, ), (1, )), ((exp1, ), (1, ))), (0.5 * v, 0.5 * v)


def _trigMul(x, y):
    if _is_const(x) or _is_const(y):
        return _mul(x, y)
    ret = _zero
    for (t1, t2), (v1, v2) in zip(product(x[0], y[0]), product(x[1], y[1])):
        v = v1 * v2
        tmp = _one
        trig = []
        for mt, n in zip(chain(t1[0], t2[0]), chain(t1[1], t2[1])):
            if mt[0] == COS:
                trig.append(mt)
            else:
                tmp = _mul(tmp, ((((mt, ), (n, )), ), (1, )))
        if len(trig) == 1:
            x = ((((trig[0], ), (1, )), ), (v, ))
            expr = _mul(tmp, x)
        elif len(trig) == 2:
            expr = _trigMul_t(trig[0], trig[1], v)
            expr = _mul(tmp, expr)
        else:
            expr = _mul(tmp, _const(v))
        ret = _add(ret, expr)
    return ret


def _exp_trig_Reduce(mtlist, v):
    trig = _one
    alpha = 0
    shift = 0
    ml, nl = [], []
    for mt, n in zip(*mtlist):
        if mt[0] == COS:
            trig = _trigMul(trig, _cos_power_n(mt, n))
        elif mt[0] == EXP:
            x = alpha * shift + n * mt[1] * mt[-1]
            alpha += n * mt[1]
            if alpha == 0:
                shift = 0
            else:
                shift = x / alpha
        elif mt[0] == GAUSSIAN and n != 1:
            ml.append((mt[0], mt[1] / np.sqrt(n), mt[2]))
            nl.append(1)
        else:
            ml.append(mt)
            nl.append(n)
    ret = (((tuple(ml), tuple(nl)), ), (v, ))

    if alpha != 0:
        ret = _mul(ret, _basic_wave(EXP, alpha, shift=shift))

    return _mul(ret, trig)


def _simplify(expr):
    ret = _zero
    for t, v in zip(*expr):
        y = _exp_trig_Reduce(t, v)
        ret = _add(ret, y)
    return ret


def _filter(expr, low, high):
    expr = _simplify(expr)
    ret = _zero
    for t, v in zip(*expr):
        for i, (mt, n) in enumerate(zip(*t)):
            if mt[0] == COS:
                if low <= mt[1] < high:
                    ret = _add(ret, ((t, ), (v, )))
                break
            elif mt[0] == SINC and n == 1:
                pass
            elif mt[0] == GAUSSIAN and n == 1:
                pass
        else:
            if low <= 0:
                ret = _add(ret, ((t, ), (v, )))
    return ret


def _apply(x, Type, shift, *args):
    return _baseFunc[Type](x - shift, *args)


def _calc(wav, x):
    lru_cache = {}

    def _calc_m(t, x):
        ret = 1
        for mt, n in zip(*t):
            if mt not in lru_cache:
                Type, *args, shift = mt
                lru_cache[mt] = _apply(x, Type, shift, *args)
            if n == 1:
                ret = ret * lru_cache[mt]
            else:
                ret = ret * lru_cache[mt]**n
        return ret

    ret = 0
    for t, v in zip(*wav):
        ret = ret + v * _calc_m(t, x)
    return ret


def _num_latex(num):
    if num == -np.inf:
        return r"-\infty"
    elif num == np.inf:
        return r"\infty"
    s = f"{num:g}"
    if "e" in s:
        a, n = s.split("e")
        n = float(n)
        s = f"{a} \\times 10^{{{n:g}}}"
    return s


def _fun_latex(fun):
    funID, *args, shift = fun
    if _baseFunc_latex[funID] is None:
        shift = _num_latex(shift)
        if shift == "0":
            shift = ""
        elif shift[0] != '-':
            shift = "+" + shift
        return r"\mathrm{Func}" + f"{funID}(t{shift}, ...)"
    return _baseFunc_latex[funID](shift, *args)


def _wav_latex(wav):
    from waveforms.waveform import _is_const, _zero

    if wav == _zero:
        return "0"
    elif _is_const(wav):
        return f"{wav[1][0]}"

    sum_expr = []
    for mul, amp in zip(*wav):
        if mul == ((), ()):
            sum_expr.append(_num_latex(amp))
            continue
        mul_expr = []
        amp = _num_latex(amp)
        if amp != "1":
            mul_expr.append(amp)
        for fun, n in zip(*mul):
            fun_expr = _fun_latex(fun)
            if n != 1:
                mul_expr.append(fun_expr + "^{" + f"{n}" + "}")
            else:
                mul_expr.append(fun_expr)
        sum_expr.append(''.join(mul_expr))

    ret = sum_expr[0]
    for expr in sum_expr[1:]:
        if expr[0] == '-':
            ret += expr
        else:
            ret += "+" + expr
    return ret


class Waveform:
    __slots__ = ('bounds', 'seq', 'max', 'min', 'start', 'stop', 'sample_rate')

    def __init__(self, bounds=(+inf, ), seq=(_zero, ), min=-inf, max=inf):
        self.bounds = bounds
        self.seq = seq
        self.max = max
        self.min = min
        self.start = None
        self.stop = None
        self.sample_rate = None

    def _head(self):
        for i, s in enumerate(self.seq):
            if s is not _zero:
                if i == 0:
                    return -inf
                return self.bounds[i - 1]
        return inf

    def _tail(self):
        N = len(self.bounds)
        for i, s in enumerate(self.seq[::-1]):
            if s is not _zero:
                if i == 0:
                    return inf
                return self.bounds[N - i - 1]
        return -inf

    @property
    def head(self):
        if self.start is None:
            return self._head()
        else:
            return max(self.start, self._head())

    @property
    def tail(self):
        if self.stop is None:
            return self._tail()
        else:
            return min(self.stop, self._tail())

    def sample(self, sample_rate=None, out=None):
        if sample_rate is None:
            sample_rate = self.sample_rate
        if self.start is None or self.stop is None or sample_rate is None:
            raise ValueError('Waveform is not initialized')
        x = np.arange(self.start, self.stop, 1 / sample_rate)
        return self.__call__(x, out=out)

    def tolist(self):
        ret = [self.max, self.min, self.start, self.stop, self.sample_rate]

        ret.append(len(self.bounds))
        for seq, b in zip(self.seq, self.bounds):
            ret.append(b)
            tlist, amplist = seq
            ret.append(len(amplist))
            for t, amp in zip(tlist, amplist):
                ret.append(amp)
                mtlist, nlist = t
                ret.append(len(nlist))
                for fun, n in zip(mtlist, nlist):
                    ret.append(n)
                    ret.append(len(fun))
                    ret.extend(fun)

        return ret

    @staticmethod
    def fromlist(l):

        def _read(l, pos, size):
            try:
                return tuple(l[pos:pos + size]), pos + size
            except:
                raise ValueError('Invalid waveform format')

        w = Waveform()
        (w.max, w.min, w.start, w.stop, w.sample_rate,
         nseg), pos = _read(l, 0, 6)
        bounds = []
        seq = []
        for _ in range(nseg):
            (b, nsum), pos = _read(l, pos, 2)
            bounds.append(b)
            amp = []
            t = []
            for _ in range(nsum):
                (a, nmul), pos = _read(l, pos, 2)
                amp.append(a)
                nlst = []
                mt = []
                for _ in range(nmul):
                    (n, nfun), pos = _read(l, pos, 2)
                    nlst.append(n)
                    fun, pos = _read(l, pos, nfun)
                    mt.append(fun)
                t.append((tuple(mt), tuple(nlst)))
            seq.append((tuple(t), tuple(amp)))
        w.seq = tuple(seq)
        w.bounds = tuple(bounds)
        return w

    def totree(self):
        header = (self.max, self.min, self.start, self.stop, self.sample_rate)
        body = []

        for seq, b in zip(self.seq, self.bounds):
            tlist, amplist = seq
            new_seq = []
            for t, amp in zip(tlist, amplist):
                mtlist, nlist = t
                new_t = []
                for fun, n in zip(mtlist, nlist):
                    new_t.append((n, fun))
                new_seq.append((amp, tuple(new_t)))
            body.append((b, tuple(new_seq)))
        return header, tuple(body)

    @staticmethod
    def fromtree(tree):
        w = Waveform()
        header, body = tree

        (w.max, w.min, w.start, w.stop, w.sample_rate) = header
        bounds = []
        seqs = []
        for b, seq in body:
            bounds.append(b)
            amp_list = []
            t_list = []
            for amp, t in seq:
                amp_list.append(amp)
                n_list = []
                mt_list = []
                for n, mt in t:
                    n_list.append(n)
                    mt_list.append(mt)
                t_list.append((tuple(mt_list), tuple(n_list)))
            seqs.append((tuple(t_list), tuple(amp_list)))
        w.bounds = tuple(bounds)
        w.seq = tuple(seqs)
        return w

    def simplify(self):
        seq = [_simplify(self.seq[0])]
        bounds = [self.bounds[0]]
        for expr, b in zip(self.seq[1:], self.bounds[1:]):
            expr = _simplify(expr)
            if expr == seq[-1]:
                seq.pop()
                bounds.pop()
            seq.append(expr)
            bounds.append(b)
        return Waveform(tuple(bounds), tuple(seq))

    def filter(self, low=0, high=inf):
        seq = []
        for expr in self.seq:
            seq.append(_filter(expr, low, high))
        return Waveform(self.bounds, tuple(seq))

    def _comb(self, other, oper):
        bounds = []
        seq = []
        i1, i2 = 0, 0
        h1, h2 = len(self.bounds), len(other.bounds)
        while i1 < h1 or i2 < h2:
            s = oper(self.seq[i1], other.seq[i2])
            b = min(self.bounds[i1], other.bounds[i2])
            if seq and s == seq[-1]:
                bounds[-1] = b
            else:
                bounds.append(b)
                seq.append(s)
            if b == self.bounds[i1]:
                i1 += 1
            if b == other.bounds[i2]:
                i2 += 1
        return Waveform(tuple(bounds), tuple(seq))

    def __pow__(self, n):
        return Waveform(self.bounds, tuple(_pow(w, n) for w in self.seq))

    def __add__(self, other):
        if isinstance(other, Waveform):
            return self._comb(other, _add)
        else:
            return self + const(other)

    def __radd__(self, v):
        return const(v) + self

    def append(self, other):
        if not isinstance(other, Waveform):
            raise TypeError('connect Waveform by other type')
        if len(self.bounds) == 1:
            self.bounds = other.bounds
            self.seq = self.seq + other.seq[1:]
            return

        assert self.bounds[-2] <= other.bounds[
            0], f"connect waveforms with overlaped domain {self.bounds}, {other.bounds}"
        if self.bounds[-2] < other.bounds[0]:
            self.bounds = self.bounds[:-1] + other.bounds
            self.seq = self.seq + other.seq[1:]
        else:
            self.bounds = self.bounds[:-2] + other.bounds
            self.seq = self.seq[:-1] + other.seq[1:]

    def __ior__(self, other):
        return self | other

    def __or__(self, other):
        if isinstance(other, (int, float, complex)):
            other = const(other)
        w = self.marker + other.marker

        def _or(a, b):
            if a != _zero or b != _zero:
                return _one
            else:
                return _zero

        return self._comb(other, _or)

    def __iand__(self, other):
        return self & other

    def __and__(self, other):
        if isinstance(other, (int, float, complex)):
            other = const(other)
        w = self.marker + other.marker

        def _and(a, b):
            if a != _zero and b != _zero:
                return _one
            else:
                return _zero

        return self._comb(other, _and)

    @property
    def marker(self):
        w = self.simplify()
        return Waveform(w.bounds,
                        tuple(_zero if s == _zero else _one for s in w.seq))

    def mask(self, edge=0):
        w = self.marker
        in_wave = w.seq[0] == _zero
        bounds = []
        seq = []

        if w.seq[0] == _zero:
            in_wave = False
            b = w.bounds[0] - edge
            bounds.append(b)
            seq.append(_zero)

        for b, s in zip(w.bounds[1:], w.seq[1:]):
            if not in_wave and s != _zero:
                in_wave = True
                bounds.append(b + edge)
                seq.append(_one)
            elif in_wave and s == _zero:
                in_wave = False
                b = b - edge
                if b > bounds[-1]:
                    bounds.append(b)
                    seq.append(_zero)
                else:
                    bounds.pop()
                    bounds.append(b)
        return Waveform(tuple(bounds), tuple(seq))

    def __mul__(self, other):
        if isinstance(other, Waveform):
            return self._comb(other, _mul)
        else:
            return self * const(other)

    def __rmul__(self, v):
        return const(v) * self

    def __truediv__(self, other):
        if isinstance(other, Waveform):
            raise TypeError('division by waveform')
        else:
            return self * const(1 / other)

    def __neg__(self):
        return -1 * self

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, v):
        return v + (-self)

    def __rshift__(self, time):
        return Waveform(
            tuple(round(bound + time, NDIGITS) for bound in self.bounds),
            tuple(_shift(expr, time) for expr in self.seq))

    def __lshift__(self, time):
        return self >> (-time)

    def __call__(self, x, frag=False, out=None):
        if isinstance(x, (int, float, complex)):
            return self.__call__(np.array([x]))[0]
        range_list = np.searchsorted(x, self.bounds)
        #ret = np.zeros_like(x)
        ret = []
        start, stop = 0, 0
        dtype = float
        for i, stop in enumerate(range_list):
            if start < stop and self.seq[i] != _zero:
                part = np.clip(_calc(self.seq[i], x[start:stop]), self.min,
                               self.max)
                if (isinstance(part, complex) or isinstance(part, np.ndarray)
                        and isinstance(part[0], complex)):
                    dtype = complex
                ret.append((start, stop, part))
            start = stop
        if not frag:
            if out is None:
                out = np.zeros_like(x, dtype=dtype)
            else:
                out *= 0
            for start, stop, part in ret:
                out[start:stop] = part
            return out
        else:
            return ret

    def __hash__(self):
        return hash((self.max, self.min, self.start, self.stop,
                     self.sample_rate, self.bounds, self.seq))

    def __eq__(self, o: object) -> bool:
        if isinstance(o, (int, float, complex)):
            return self == const(o)
        elif isinstance(o, Waveform):
            a = self.simplify()
            b = o.simplify()
            return a.seq == b.seq and a.bounds == b.bounds and (
                a.max, a.min, a.start, a.stop,
                a.sample_rate) == (b.max, b.min, b.start, b.stop,
                                   b.sample_rate)
        else:
            return False

    def _repr_latex_(self):
        parts = []
        start = -np.inf
        for end, wav in zip(self.bounds, self.seq):
            e_str = _wav_latex(wav)
            start_str = _num_latex(start)
            end_str = _num_latex(end)
            parts.append(e_str + r",~~&t\in" + f"({start_str},{end_str}" +
                         (']' if end < np.inf else ')'))
            start = end
        if len(parts) == 1:
            expr = ''.join(['f(t)=', *parts[0].split('&')])
        else:
            expr = '\n'.join([
                r"f(t)=\begin{cases}", (r"\\" + '\n').join(parts),
                r"\end{cases}"
            ])
        return "$$\n{}\n$$".format(expr)


_zero_waveform = Waveform()
_one_waveform = Waveform(seq=(_one, ))


def zero():
    return _zero_waveform


def one():
    return _one_waveform


def const(c):
    return Waveform(seq=(_const(1.0 * c), ))


__TypeIndex = 1
_baseFunc = {}
_derivativeBaseFunc = {}
_baseFunc_latex = {}


def registerBaseFunc(func, latex=None):
    global __TypeIndex
    Type = __TypeIndex
    __TypeIndex += 1

    _baseFunc[Type] = func
    _baseFunc_latex[Type] = latex

    return Type


def packBaseFunc():
    return pickle.dumps(_baseFunc)


def updateBaseFunc(buf):
    _baseFunc.update(pickle.loads(buf))


def registerDerivative(Type, dFunc):
    _derivativeBaseFunc[Type] = dFunc


# register base function
def _format_LINEAR(shift, *args):
    if shift != 0:
        shift = _num_latex(-shift)
        if shift[0] == '-':
            return f"(t{shift})"
        else:
            return f"(t+{shift})"
    else:
        return 't'


def _format_GAUSSIAN(shift, *args):
    sigma = _num_latex(args[0] / np.sqrt(2))
    shift = _num_latex(-shift)
    if shift != '0':
        if shift[0] != '-':
            shift = '+' + shift
        if sigma == '1':
            return ('\\exp\\left[-\\frac{\\left(t' + shift +
                    '\\right)^2}{2}\\right]')
        else:
            return ('\\exp\\left[-\\frac{1}{2}\\left(\\frac{t' + shift + '}{' +
                    sigma + '}\\right)^2\\right]')
    else:
        if sigma == '1':
            return ('\\exp\\left(-\\frac{t^2}{2}\\right)')
        else:
            return ('\\exp\\left[-\\frac{1}{2}\\left(\\frac{t}{' + sigma +
                    '}\\right)^2\\right]')


def _format_SINC(shift, *args):
    shift = _num_latex(-shift)
    bw = _num_latex(args[0])
    if shift != '0':
        if shift[0] != '-':
            shift = '+' + shift
        if bw == '1':
            return '\\mathrm{sinc}(t' + shift + ')'
        else:
            return '\\mathrm{sinc}[' + bw + '(t' + shift + ')]'
    else:
        if bw == '1':
            return '\\mathrm{sinc}(t)'
        else:
            return '\\mathrm{sinc}(' + bw + 't)'


def _format_COSINE(shift, *args):
    freq = args[0] / 2 / np.pi
    phase = -shift * freq
    freq = _num_latex(freq)
    if freq == '1':
        freq = ''
    phase = _num_latex(phase)
    if phase == '0':
        phase = ''
    elif phase[0] != '-':
        phase = '+' + phase
    if phase != '':
        return f'\\cos\\left[2\\pi({freq}t{phase})\\right]'
    elif freq != '':
        return f'\\cos\\left(2\\pi\\times {freq}t\\right)'
    else:
        return '\\cos(2\\pi t)'


def _format_ERF(shift, *args):
    if shift != 0:
        return '\\mathrm{erf}(\\frac{t-' + f"{shift:g}" + '}{' + f'{args[0]:g}' + '})'
    else:
        return '\\mathrm{erf}(\\frac{t}{' + f'{args[0]:g}' + '})'


def _format_EXP(shift, *args):
    if shift != 0:
        return '\\exp(-' + f'{args[0]:g}' + '(t-' + f"{shift:g}" + '))'
    else:
        return '\\exp(-' + f'{args[0]:g}' + 't)'


LINEAR = registerBaseFunc(lambda t: t, _format_LINEAR)
GAUSSIAN = registerBaseFunc(lambda t, std_sq2: np.exp(-(t / std_sq2)**2),
                            _format_GAUSSIAN)
ERF = registerBaseFunc(lambda t, std_sq2: special.erf(t / std_sq2),
                       _format_ERF)
COS = registerBaseFunc(lambda t, w: np.cos(w * t), _format_COSINE)
SINC = registerBaseFunc(lambda t, bw: np.sinc(bw * t), _format_SINC)
EXP = registerBaseFunc(lambda t, alpha: np.exp(alpha * t), _format_EXP)
INTERP = registerBaseFunc(lambda t, start, stop, points: np.interp(
    t, np.linspace(start, stop, len(points)), points))
LINEARCHIRP = registerBaseFunc(lambda t, f0, f1, T, phi0: np.sin(
    phi0 + 2 * np.pi * ((f1 - f0) / (2 * T) * t**2 + f0 * t)))
EXPONENTIALCHIRP = registerBaseFunc(lambda t, f0, alpha, phi0: np.sin(
    phi0 + 2 * pi * f0 * (np.exp(alpha * t) - 1) / alpha))
HYPERBOLICCHIRP = registerBaseFunc(lambda t, f0, k, phi0: np.sin(
    phi0 + 2 * np.pi * f0 / k * np.log(1 + k * t)))

# register derivative
registerDerivative(LINEAR, lambda shift, *args: _one)

registerDerivative(
    GAUSSIAN, lambda shift, *args: (((((LINEAR, shift),
                                       (GAUSSIAN, *args, shift)), (1, 1)), ),
                                    (-2 / args[0]**2, )))

registerDerivative(
    ERF, lambda shift, *args: (((((GAUSSIAN, *args, shift), ), (1, )), ),
                               (2 / args[0] / np.sqrt(pi), )))

registerDerivative(
    COS, lambda shift, *args: (((((COS, args[0], shift - pi / args[0] / 2), ),
                                 (1, )), ), (args[0], )))

registerDerivative(
    SINC, lambda shift, *args:
    (((((LINEAR, shift), (COS, *args, shift)), (-1, 1)),
      (((LINEAR, shift), (COS, args[0], args[1] - pi / 2, shift)), (-2, 1))),
     (1, -1 / args[0])))

registerDerivative(
    EXP, lambda shift, *args: (((((EXP, *args, shift), ), (1, )), ),
                               (args[0], )))

registerDerivative(
    INTERP, lambda shift, start, stop, points:
    (((((INTERP, start, stop, tuple(np.gradient(np.asarray(points))), shift),
        ), (1, )), ), ((len(points) - 1) / (stop - start), )))


def _d_LINEARCHIRP(shift, f0, f1, T, phi0):
    tlist = (
        (((LINEARCHIRP, f0, f1, T, phi0 + pi / 2, shift), ), (1, )),
        (((LINEAR, shift), (LINEARCHIRP, f0, f1, T, phi0 + pi / 2, shift)),
         (1, 1)),
    )
    alist = (2 * pi * f0, 2 * pi * (f1 - f0) / T)

    if f0 == 0:
        return tlist[1:], alist[1:]
    else:
        return tlist, alist


registerDerivative(LINEARCHIRP, _d_LINEARCHIRP)
registerDerivative(
    EXPONENTIALCHIRP, lambda shift, f0, alpha, phi0:
    (((((EXP, alpha, shift),
        (EXPONENTIALCHIRP, f0, alpha, phi0 + pi / 2, shift)), (1, 1)), ),
     (2 * pi * f0, )))
registerDerivative(
    HYPERBOLICCHIRP, lambda shift, f0, k, phi0:
    (((((LINEAR, shift - 1 / k),
        (HYPERBOLICCHIRP, f0, k, phi0 + pi / 2, shift)), (-1, 1)), ),
     (2 * pi * f0, )))


def _D_base(m):
    Type, *args, shift = m
    return _derivativeBaseFunc[Type](shift, *args)


def _D(x):
    if _is_const(x):
        return _zero
    t_list, v_list = x
    if len(v_list) == 1:
        (m_list, n_list), v = t_list[0], v_list[0]
        if len(m_list) == 1:
            m, n = m_list[0], n_list[0]
            if n == 1:
                return _mul(_D_base(m), _const(v))
            else:
                return _mul(((((m, ), (n - 1, )), ), (n * v, )),
                            _D(((((m, ), (1, )), ), (1, ))))
        else:
            a = (((m_list[:1], n_list[:1]), ), (v, ))
            b = (((m_list[1:], n_list[1:]), ), (1, ))
            return _add(_mul(a, _D(b)), _mul(_D(a), b))
    else:
        return _add(_D((t_list[:1], v_list[:1])), _D((t_list[1:], v_list[1:])))


def D(wav):
    """derivative
    """
    return Waveform(bounds=wav.bounds, seq=tuple(_D(x) for x in wav.seq))


def convolve(a, b):
    pass


def sign():
    return Waveform(bounds=(0, +inf), seq=(_const(-1), _one))


def step(edge, type='erf'):
    """
    type: "erf", "cos", "linear"
    """
    if edge == 0:
        return Waveform(bounds=(0, +inf), seq=(_zero, _one))
    if type == 'cos':
        rise = _add(_half,
                    _mul(_half, _basic_wave(COS, pi / edge, shift=0.5 * edge)))
        return Waveform(bounds=(round(-edge / 2,
                                      NDIGITS), round(edge / 2,
                                                      NDIGITS), +inf),
                        seq=(_zero, rise, _one))
    elif type == 'linear':
        rise = _add(_half, _mul(_const(1 / edge), _basic_wave(LINEAR)))
        return Waveform(bounds=(round(-edge / 2,
                                      NDIGITS), round(edge / 2,
                                                      NDIGITS), +inf),
                        seq=(_zero, rise, _one))
    else:
        std_sq2 = edge / 5
        # rise = _add(_half, _mul(_half, _basic_wave(ERF, std_sq2)))
        rise = ((((), ()), (((ERF, std_sq2, 0), ), (1, ))), (0.5, 0.5))
        return Waveform(bounds=(-round(edge, NDIGITS), round(edge,
                                                             NDIGITS), +inf),
                        seq=(_zero, rise, _one))


def square(width, edge=0, type='erf'):
    if width <= 0:
        return zero()
    if edge == 0:
        return Waveform(bounds=(round(-0.5 * width,
                                      NDIGITS), round(0.5 * width,
                                                      NDIGITS), +inf),
                        seq=(_zero, _one, _zero))
    else:
        return ((step(edge, type=type) << width / 2) -
                (step(edge, type=type) >> width / 2))


def gaussian(width):
    if width <= 0:
        return zero()
    # width is two times FWHM
    # std_sq2 = width / (4 * np.sqrt(np.log(2)))
    std_sq2 = width / 3.3302184446307908
    # std is set to give total pulse area same as a square
    #std_sq2 = width/np.sqrt(np.pi)
    return Waveform(bounds=(round(-0.75 * width,
                                  NDIGITS), round(0.75 * width,
                                                  NDIGITS), +inf),
                    seq=(_zero, _basic_wave(GAUSSIAN, std_sq2), _zero))


def cos(w, phi=0):
    if w == 0:
        return const(np.cos(phi))
    if w < 0:
        phi = -phi
        w = -w
    return Waveform(seq=(_basic_wave(COS, w, shift=-phi / w), ))


def sin(w, phi=0):
    if w == 0:
        return const(np.sin(phi))
    if w < 0:
        phi = -phi + pi
        w = -w
    return Waveform(seq=(_basic_wave(COS, w, shift=(pi / 2 - phi) / w), ))


def exp(alpha):
    if isinstance(alpha, complex):
        if alpha.real == 0:
            return cos(alpha.imag) + 1j * sin(alpha.imag)
        else:
            return exp(alpha.real) * (cos(alpha.imag) + 1j * sin(alpha.imag))
    else:
        return Waveform(seq=(_basic_wave(EXP, alpha), ))


def sinc(bw):
    if bw <= 0:
        return zero()
    width = 100 / bw
    return Waveform(bounds=(round(-0.5 * width,
                                  NDIGITS), round(0.5 * width, NDIGITS), +inf),
                    seq=(_zero, _basic_wave(SINC, bw), _zero))


def cosPulse(width):
    # cos = _basic_wave(COS, 2*np.pi/width)
    # pulse = _mul(_add(cos, _one), _half)
    if width <= 0:
        return zero()
    pulse = ((((), ()), (((COS, 6.283185307179586 / width, 0), ), (1, ))),
             (0.5, 0.5))
    return Waveform(bounds=(round(-0.5 * width,
                                  NDIGITS), round(0.5 * width, NDIGITS), +inf),
                    seq=(_zero, pulse, _zero))


def hanning(width):
    return cosPulse(width)


def general_cosine(duration, *arg):
    wav = zero()
    arg = np.asarray(arg)
    arg /= arg[::2].sum()
    for i, a in enumerate(arg, start=1):
        wav += a / 2 * (1 - (-1)**i * cos(i * 2 * pi / duration))
    return wav * square(duration)


def slepian(duration, *arg):
    wav = zero()
    arg = np.asarray(arg)
    arg /= arg[::2].sum()
    for i, a in enumerate(arg, start=1):
        wav += a / 2 * (1 - (-1)**i * cos(i * 2 * pi / duration))
    return wav * square(duration)


def _poly(*a):
    """
    a[0] + a[1] * t + a[2] * t**2 + ...
    """
    t = (((), ()), *[(((LINEAR, 0), ), (n, ))
                     for n, _ in enumerate(a[1:], start=1)])
    return t, a


def poly(a):
    """
    a[0] + a[1] * t + a[2] * t**2 + ...
    """
    return Waveform(seq=(_poly(*a), ))


def chirp(f0, f1, T, phi0=0, type='linear'):
    """
    A chirp is a signal in which the frequency increases (up-chirp)
    or decreases (down-chirp) with time. In some sources, the term
    chirp is used interchangeably with sweep signal.

    type: "linear", "exponential", "hyperbolic"
    """
    if f0 == f1:
        return sin(f0, phi0)
    if T <= 0:
        raise ValueError('T must be positive')

    if type == 'linear':
        # f(t) = f1 * (t/T) + f0 * (1 - t/T)
        return Waveform(bounds=(0, round(T, NDIGITS), +inf),
                        seq=(_zero, _basic_wave(LINEARCHIRP, f0, f1, T,
                                                phi0), _zero))
    elif type in ['exp', 'exponential', 'geometric']:
        # f(t) = f0 * (f1/f0) ** (t/T)
        if f0 == 0:
            raise ValueError('f0 must be non-zero')
        alpha = np.log(f1 / f0) / T
        return Waveform(bounds=(0, round(T, NDIGITS), +inf),
                        seq=(_zero,
                             _basic_wave(EXPONENTIALCHIRP, f0, alpha,
                                         phi0), _zero))
    elif type in ['hyperbolic', 'hyp']:
        # f(t) = f0 * f1 / (f0 * (t/T) + f1 * (1-t/T))
        if f0 * f1 == 0:
            return const(np.sin(phi0))
        k = (f0 - f1) / (f1 * T)
        return Waveform(bounds=(0, round(T, NDIGITS), +inf),
                        seq=(_zero, _basic_wave(HYPERBOLICCHIRP, f0, k,
                                                phi0), _zero))
    else:
        raise ValueError(f'unknown type {type}')


def interp(x, y):
    seq, bounds = [_zero], [x[0]]
    for x1, x2, y1, y2 in zip(x[:-1], x[1:], y[:-1], y[1:]):
        if x2 == x1:
            continue
        seq.append(
            _add(
                _mul(_const((y2 - y1) / (x2 - x1)),
                     _basic_wave(LINEAR, shift=x1)), _const(y1)))
        bounds.append(x2)
    bounds.append(inf)
    seq.append(_zero)
    return Waveform(seq=tuple(seq),
                    bounds=tuple(round(b, NDIGITS)
                                 for b in bounds)).simplify()


def cut(wav, start=None, stop=None, head=None, tail=None, min=None, max=None):
    offset = 0
    if start is not None and head is not None:
        offset = head - wav(np.array([1.0 * start]))[0]
    elif stop is not None and tail is not None:
        offset = tail - wav(np.array([1.0 * stop]))[0]
    wav = wav + offset

    if start is not None:
        wav = wav * (step(0) >> start)
    if stop is not None:
        wav = wav * ((1 - step(0)) >> stop)
    if min is not None:
        wav.min = min
    if max is not None:
        wav.max = max
    return wav


def function(fun, *args, start=None, stop=None):
    TYPEID = registerBaseFunc(fun)
    seq = (_basic_wave(TYPEID, *args), )
    wav = Waveform(seq=seq)
    if start is not None:
        wav = wav * (step(0) >> start)
    if stop is not None:
        wav = wav * ((1 - step(0)) >> stop)
    return wav


def samplingPoints(start, stop, points):
    return Waveform(bounds=(round(start, NDIGITS), round(stop, NDIGITS), inf),
                    seq=(_zero, _basic_wave(INTERP, start, stop,
                                            tuple(points)), _zero))


def mixing(I,
           Q=None,
           *,
           phase=0.0,
           freq=0.0,
           ratioIQ=1.0,
           phaseDiff=0.0,
           block_freq=None,
           DRAGScaling=None):
    """SSB or envelope mixing
    """
    if Q is None:
        I = I
        Q = zero()

    w = 2 * pi * freq
    if freq != 0.0:
        # SSB mixing
        Iout = I * cos(w, -phase) + Q * sin(w, -phase)
        Qout = -I * sin(w, -phase + phaseDiff) + Q * cos(w, -phase + phaseDiff)
    else:
        # envelope mixing
        Iout = I * np.cos(-phase) + Q * np.sin(-phase)
        Qout = -I * np.sin(-phase) + Q * np.cos(-phase)

    # apply DRAG
    if block_freq is not None:
        a = block_freq / (block_freq - freq)
        b = 1 / (block_freq - freq)
        I = a * Iout + b / (2 * pi) * D(Qout)
        Q = a * Qout - b / (2 * pi) * D(Iout)
        Iout, Qout = I, Q
    elif DRAGScaling is not None and DRAGScaling != 0:
        I = (1 - w * DRAGScaling) * Iout - DRAGScaling * D(Qout)
        Q = (1 - w * DRAGScaling) * Qout + DRAGScaling * D(Iout)
        Iout, Qout = I, Q

    Qout = ratioIQ * Qout

    return Iout, Qout


__all__ = [
    'D', 'Waveform', 'chirp', 'const', 'cos', 'cosPulse', 'cut', 'exp',
    'function', 'gaussian', 'general_cosine', 'hanning', 'interp', 'mixing',
    'one', 'poly', 'registerBaseFunc', 'registerDerivative', 'samplingPoints',
    'sign', 'sin', 'sinc', 'square', 'step', 'zero'
]



import importlib
import io
import logging
import sys

import git

log = logging.getLogger(__name__)


class GitFinder(importlib.abc.MetaPathFinder):
    def __init__(self, path, name='HEAD'):
        self.path = path
        self.name = name

    def find_spec(self, fullname, path=None, target=None):
        log.debug("find_spec: fullname=%r, path=%r, target=%r", fullname, path,
                  target)
        loader = GitModuleLoader(self.path, self.name, fullname)
        if loader.blob is not None:
            log.debug("find_spec: module %r found", fullname)
            return importlib.util.spec_from_loader(fullname, loader)
        else:
            log.debug("find_spec: module %r not found", fullname)
            return None


class GitModuleLoader(importlib.abc.FileLoader, importlib.abc.SourceLoader):
    def __init__(self, repo_path, name, fullname):
        self.repo_path = repo_path
        self.hexsha = git.Repo(self.repo_path).commit(name).hexsha

        self.blob, self.filepath, self._is_package = self._find_blob(
            repo_path, name, '/'.join(fullname.split('.')))

    def _find_blob(self, repo_path, name, filepath):
        try:
            # find a module
            blob = git.Repo(repo_path).commit(name).tree[filepath + '.py']
            return blob, filepath + '.py', False
        except KeyError:
            pass
        try:
            # find a package
            blob = git.Repo(repo_path).commit(name).tree[filepath +
                                                         '/__init__.py']
            return blob, filepath + '/__init__.py', True
        except KeyError:
            pass
        return None, None, False

    def is_package(self, fullname):
        return self._is_package

    def get_filename(self, fullname):
        return f"git://{self.hexsha}@{self.repo_path}|{self.filepath}"

    def get_data(self, path):
        buf = io.BytesIO()
        self.blob.stream_data(buf)
        buf.seek(0)
        return buf.read()


_installed_meta_cache = {}


def install_meta(path, name='HEAD'):
    if (path, name) not in _installed_meta_cache:
        finder = GitFinder(path, name)
        _installed_meta_cache[(path, name)] = finder
        for i, v in enumerate(sys.meta_path):
            if isinstance(v, GitFinder):
                sys.meta_path.insert(i, finder)
                break
        else:
            sys.meta_path.append(finder)
        log.debug('%r installed on sys.meta_path', finder)


def remove_meta(path, name='HEAD'):
    if (path, name) in _installed_meta_cache:
        finder = _installed_meta_cache.pop((path, name))
        sys.meta_path.remove(finder)
        log.debug('%r removed from sys.meta_path', finder)



import numpy as np
from waveforms.scan_iter import *
import pytest


class FindPeak(BaseOptimizer):

    def __init__(self, dimensions):
        self.index = -1
        self.max_args = dimensions[0]
        self.max_value = -np.inf

    def tell(self, suggested, value):
        args, value = value
        if self.max_value <= value:
            self.max_value = value
            self.max_args = args

    def ask(self):
        self.index += 1
        self.max_value = -np.inf
        return self.max_args,

    def get_result(self):

        class Result():
            x = self.max_args,

        return Result()


def spectrum_func(b, f):
    c = 1 - b**2
    return np.exp(-((f - c) / 0.01)**2)


def spectrum_mask(freq, center=None):
    if center is None:
        return True
    else:
        return -0.1 <= freq - center <= 0.1


@pytest.fixture
def spectrum_data():
    z = np.full((101, 121), np.nan)
    iq = np.full((101, 121, 1024), np.nan, dtype=complex)
    obj = np.full((101, 121), np.nan, dtype=object)
    center = None
    bias_list = np.linspace(-0.1, 0.1, 101)
    freq_list = np.linspace(-0.1, 1.1, 121)

    np.random.seed(1234)
    for i, bias in enumerate(bias_list):
        for j, freq in enumerate(freq_list):
            if spectrum_mask(freq, center):
                z[i, j] = spectrum_func(bias, freq)
                iq[i,
                   j, :] = np.random.randn(1024) + 1j * np.random.randn(1024)
                obj[i, j] = {'a': 1, 'b': 2}
        center = freq_list[np.argmin(np.abs(freq_list - (1 - bias**2)))]
    return {
        'z': z,
        'iq': iq,
        'obj': obj,
        'bias_list': bias_list,
        'freq_list': freq_list
    }


def test_scan_iter():

    def f(a):
        for i in range(5):
            yield a + i, i

    def f2():
        for i in range(2):
            yield i

    def f3():
        for i in range(2):
            yield i * 100, -i * 200

    steps = scan_iters(
        {
            'a': [-1, 1],
            ('b', ('c', 'd')): ([4, 5, 6], f),
            (('e', ), ('g', 'h')): (f2, f3)
        },
        filter=lambda x: x < 8,
        functions={
            'x': lambda a, b: a + b
        })

    def scan_iter2():
        for a in [-1, 1]:
            for b, (c, d) in zip([4, 5, 6], f(a)):
                for e, (g, h) in zip(f2(), f3()):
                    x = a + b
                    if x < 8:
                        yield {
                            'a': a,
                            'b': b,
                            'c': c,
                            'd': d,
                            'e': e,
                            'g': g,
                            'h': h,
                            'x': x
                        }

    for step, args in zip(steps, scan_iter2()):
        for k in ['a', 'b', 'c', 'd', 'e', 'g', 'h', 'x']:
            assert k in step.kwds
            assert step.kwds[k] == args[k]


def test_scan_iter2():

    def gen(a, b, x, y):
        for i in range(a):
            yield b + x + y

    def scan_iter2():
        for a in [1, 2, 3]:
            x = a + 0.5
            for b in x * np.arange(3):
                y = x + a + b
                for c in gen(a, b, x, y):
                    yield {'a': a, 'b': b, 'c': c, 'x': x, 'y': y}

    info = {
        'loops': {
            'a': [1, 2, 3],
            'b': lambda x: x * np.arange(3),
            'c': gen
        },
        'functions': {
            'x': lambda a: a + 0.5,
            'y': lambda a, b, x: x + a + b
        }
    }

    for step, args in zip(scan_iters(**info), scan_iter2()):
        for k in ['a', 'b', 'c', 'x', 'y']:
            assert k in step.kwds
            assert step.kwds[k] == args[k]


def test_scan_iter3():

    N = 3

    def scan_iter2():
        for x0, x1, x2 in zip(*[np.arange(5) + 5 * i for i in range(N)]):
            z0, z1, z2 = x0, x1, x2
            for y0, y1, y2 in zip(*[
                    x0 * np.array([-1, 1]), x1 * np.array([-1, 1]), x2 *
                    np.array([-1, 1])
            ]):
                yield {
                    'x0': x0,
                    'x1': x1,
                    'x2': x2,
                    'y0': y0,
                    'y1': y1,
                    'y2': y2,
                    'z0': z0,
                    'z1': z1,
                    'z2': z2
                }

    info = {
        'loops': {
            tuple([f"x{i}" for i in range(N)]):
            tuple([np.arange(5) + 5 * i for i in range(N)]),
            tuple([f"y{i}" for i in range(N)]):
            tuple([
                lambda i=i, **kw: kw[f"x{i}"] * np.array([-1, 1])
                for i in range(N)
            ]),
        },
        'functions':
        {f"z{i}": lambda i=i, **kw: kw[f"x{i}"]
         for i in range(N)}
    }
    for step, args in zip(scan_iters(**info), scan_iter2()):
        for k in ['x0', 'x1', 'x2', 'y0', 'y1', 'y2', 'z0', 'z1', 'z2']:
            assert k in step.kwds
            assert step.kwds[k] == args[k]


def test_scan_iter4():
    return

    def gen(a, b, x, y):
        for i in range(a):
            yield b + x + y

    def scan_iter2():
        for a0, a1 in zip([1, 2, 3], [4, 5, 6]):
            x0 = a0 + 0.5
            x1 = a1 + 0.5
            for b0, b1 in zip(x0 * np.arange(3), x1 * np.arange(3)):
                y0 = x0 + a0 + b0
                y1 = x1 + a1 + b1
                for c0, c1 in zip(gen(a0, b0, x0, y0), gen(a1, b1, x1, y1)):
                    yield {
                        'a0': a0,
                        'b0': b0,
                        'c0': c0,
                        'x0': x0,
                        'y0': y0,
                        'a1': a1,
                        'b1': b1,
                        'c1': c1,
                        'x1': x1,
                        'y1': y1
                    }

    info = {
        'loops': {
            tuple([f'a{i}' for i in range(2)]): ([1, 2, 3], [4, 5, 6]),
            tuple([f'b{i}' for i in range(2)]):
            tuple([
                lambda i=i, **kw: kw[f"x{i}"] * np.arange(3) for i in range(2)
            ]),
            tuple([f'c{i}' for i in range(2)]):
            gen
        },
        'functions': {
            tuple([f'x{i}' for i in range(2)]): lambda a: a + 0.5,
            tuple([f'y{i}' for i in range(2)]): lambda a, b, x: x + a + b
        }
    }

    for step, args in zip(scan_iters(**info), scan_iter2()):
        for k in ['a', 'b', 'c', 'x', 'y']:
            assert k in step.kwds
            assert step.kwds[k] == args[k]


def test_storage(spectrum_data):
    z = spectrum_data['z']
    iq = spectrum_data['iq']
    bias_list = spectrum_data['bias_list']
    freq_list = spectrum_data['freq_list']

    data1 = Storage(save_kwds=False)
    data2 = Storage(save_kwds=True)

    np.random.seed(1234)
    for step in scan_iters(
        {
            ('bias', 'center'):
            (bias_list, OptimizerConfig(FindPeak, [None], max_iters=101)),
            'freq':
            freq_list,
        },
            filter=spectrum_mask,
            trackers=[data1, data2]):
        y = spectrum_func(step.kwds['bias'], step.kwds['freq'])

        step.store({
            'z': y,
            'iq': np.random.randn(1024) + 1j * np.random.randn(1024),
            'obj': {
                'a': 1,
                'b': 2
            }
        })
        step.feedback(('center', ), (step.kwds['freq'], y))

    for data in [data1, data2]:
        assert np.all(bias_list == data['bias'])
        assert np.all(freq_list == data['freq'])
        assert data['z'].shape == (101, 121)
        assert data['iq'].shape == (101, 121, 1024)
        assert np.all((z == data['z'])[np.isnan(z) == False])
        assert np.all((iq == data['iq'])[np.isnan(iq) == False])

    assert set(data1.keys()) == {'bias', 'freq', 'z', 'iq', 'obj'}

    assert set(data2.keys()) == {'bias', 'freq', 'z', 'iq', 'obj', 'center'}
    assert data2['center'].shape == (101, 121)

    for key in ['z', 'iq', 'obj']:
        assert np.allclose(data1.timestamps[key], data2.timestamps[key])


def test_storage_future(spectrum_data):
    from concurrent.futures import ThreadPoolExecutor

    def init_thread():
        np.random.seed(1234)

    def get_result(bias, freq):
        y = spectrum_func(bias, freq)
        return {
            'z': y,
            'iq': np.random.randn(1024) + 1j * np.random.randn(1024),
            'obj': {
                'a': 1,
                'b': 2
            }
        }

    pool = ThreadPoolExecutor(max_workers=1, initializer=init_thread)

    z = spectrum_data['z']
    iq = spectrum_data['iq']
    bias_list = spectrum_data['bias_list']
    freq_list = spectrum_data['freq_list']

    data = Storage(save_kwds=False)

    for step in scan_iters(
        {
            ('bias', 'center'):
            (bias_list, OptimizerConfig(FindPeak, [None], max_iters=101)),
            'freq':
            freq_list,
        },
            filter=spectrum_mask,
            trackers=[data]):

        fut = pool.submit(get_result, step.kwds['bias'], step.kwds['freq'])

        step.store(fut)
        step.feedback(('center', ), (step.kwds['freq'], fut.result()['z']))

    assert set(data.keys()) == {'bias', 'freq', 'z', 'iq', 'obj'}
    assert np.all(bias_list == data['bias'])
    assert np.all(freq_list == data['freq'])
    assert data['z'].shape == (101, 121)
    assert data['iq'].shape == (101, 121, 1024)
    assert data['obj'].shape == (101, 121)
    assert data['obj'][0, 0] == {'a': 1, 'b': 2}
    assert np.all((z == data['z'])[np.isnan(z) == False])
    assert np.all((iq == data['iq'])[np.isnan(iq) == False])

    pool.shutdown()


def test_level_marker():
    iters = {'a': range(2), 'b': range(2), 'c': range(2)}

    def scan_iter2():
        for a in range(2):
            yield {'a': a}, 0, 'begin'
            for b in range(2):
                yield {'a': a, 'b': b}, 1, 'begin'
                for c in range(2):
                    yield {'a': a, 'b': b, 'c': c}, 2, 'begin'
                    yield {'a': a, 'b': b, 'c': c}, 2, 'step'
                    yield {'a': a, 'b': b, 'c': c}, 2, 'end'
                yield {'a': a, 'b': b}, 1, 'end'
            yield {'a': a}, 0, 'end'

    for step, args in zip(scan_iters(iters, level_marker=True), scan_iter2()):
        kw, level, marker = args
        assert step.kwds == kw
        if marker == 'begin':
            assert isinstance(step, Begin)
            assert step.level == level
        elif marker == 'step':
            assert isinstance(step, StepStatus)
        elif marker == 'end':
            assert isinstance(step, End)
            assert step.level == level



import numpy as np
import scipy.sparse as sp


def issparse(qob):
    """Checks if ``qob`` is explicitly sparse.
    """
    return isinstance(qob, sp.spmatrix)


def make_immutable(mat):
    """Make array read only, in-place.
    Parameters
    ----------
    mat : sparse or dense array
        Matrix to make immutable.
    """
    if issparse(mat):
        mat.data.flags.writeable = False
        if mat.format in {'csr', 'csc', 'bsr'}:
            mat.indices.flags.writeable = False
            mat.indptr.flags.writeable = False
        elif mat.format == 'coo':
            mat.row.flags.writeable = False
            mat.col.flags.writeable = False
    else:
        mat.flags.writeable = False
    return mat


# Paulis
sigmaI = lambda: make_immutable(np.eye(2, dtype=complex))
sigmaX = lambda: make_immutable(np.array([[0, 1], [1, 0]], dtype=complex))
sigmaY = lambda: make_immutable(np.array([[0, -1j], [1j, 0]], dtype=complex))
sigmaZ = lambda: make_immutable(np.array([[1, 0], [0, -1]], dtype=complex))

# Bell states
BellPhiP = lambda: make_immutable(
    np.array([1, 0, 0, 1], dtype=complex) / np.sqrt(2))
BellPhiM = lambda: make_immutable(
    np.array([1, 0, 0, -1], dtype=complex) / np.sqrt(2))
BellPsiP = lambda: make_immutable(
    np.array([0, 1, 1, 0], dtype=complex) / np.sqrt(2))
BellPsiM = lambda: make_immutable(
    np.array([0, 1, -1, 0], dtype=complex) / np.sqrt(2))
phiplus, phiminus = BellPhiP(), BellPhiM()
psiplus, psiminus = BellPsiP(), BellPsiM()

##########################################################


def U(theta, phi, lambda_, delta=0):
    """general unitary
    
    Any general unitary could be implemented in 2 pi/2-pulses on hardware

    U(theta, phi, lambda_, delta) = \
        np.exp(1j * delta) * \
        U(0, 0, theta + phi + lambda_) @ \
        U(np.pi / 2, p2, -p2) @ \
        U(np.pi / 2, p1, -p1))

    or  = \
        np.exp(1j * delta) * \
        U(0, 0, theta + phi + lambda_) @ \
        rfUnitary(np.pi / 2, p2 + pi / 2) @ \
        rfUnitary(np.pi / 2, p1 + pi / 2)
    
    where p1 = -lambda_ - pi / 2
          p2 = pi / 2 - theta - lambda_
    """
    c, s = np.cos(theta / 2), np.sin(theta / 2)
    a, b = (phi + lambda_) / 2, (phi - lambda_) / 2
    d = np.exp(1j * delta)
    return d * np.array([[c * np.exp(-1j * a), -s * np.exp(-1j * b)],
                         [s * np.exp(1j * b), c * np.exp(1j * a)]])


def Unitary2Angles(U: np.ndarray) -> np.ndarray:
    if U[0, 0] == 0:
        delta = (np.angle(U[1, 0]) + np.angle(U[0, 1])) / 2
        U /= np.exp(1j * delta)
        theta = np.pi
        phi = np.angle(U[1, 0])
        lambda_ = -phi
    else:
        delta = np.angle(U[0, 0])
        U = U / np.exp(1j * delta)
        theta = 2 * np.arccos(U[0, 0])
        phi = np.angle(U[1, 0])
        lambda_ = np.angle(-U[0, 1])
        delta += (phi + lambda_) / 2
    return np.array([theta, phi, lambda_, delta]).real


def rfUnitary(theta, phi):
    """
    Gives the unitary operator for an ideal microwave gate.
    phi gives the rotation axis on the plane of the bloch sphere (RF drive phase)
    theta is the rotation angle of the gate (pulse area)

    rfUnitary(theta, phi) := expm(-1j * theta / 2 * \
        (sigmaX() * cos(phi) + sigmaY() * sin(phi)))

    rfUnitary(theta, phi + pi/2) == U(theta, phi, -phi)
    """
    c, s = np.cos(theta / 2), np.sin(theta / 2)
    return np.array([[c, -1j * s * np.exp(-1j * phi)],
                     [-1j * s * np.exp(1j * phi), c]])


def fSim(theta, phi):
    c, s = np.cos(theta), np.sin(theta)
    p = np.exp(-1j * phi)
    return np.array([
        [1,     0,     0,     0],
        [0,     c, -1j*s,     0],
        [0, -1j*s,     c,     0],
        [0,     0,     0,     p]
    ]) #yapf: disable
